@@ -65,26 +65,27 @@ struct KeyPairPaths {
     QString pub;
 };
 
-QString gpgExecutable()
+bool waitOrTerminate(QProcess *p, int timeoutMs)
 {
-    return "gpg2";
+    p->waitForStarted();
+
+    if ( p->state() != QProcess::NotRunning && !p->waitForFinished(timeoutMs) ) {
+        p->terminate();
+        if ( !p->waitForFinished(5000) )
+            p->kill();
+        return false;
+    }
+
+    return true;
 }
 
-QStringList getDefaultEncryptCommandArguments(const QString &publicKeyPath)
+bool verifyProcess(QProcess *p, int timeoutMs = 30000)
 {
-    return QStringList() << "--trust-model" << "always" << "--recipient" << "copyq"
-                         << "--charset" << "utf-8" << "--display-charset" << "utf-8" << "--no-tty"
-                         << "--no-default-keyring" << "--keyring" << publicKeyPath;
-}
+    if ( !waitOrTerminate(p, timeoutMs) ) {
+        log( "ItemEncrypt ERROR: Process timed out; stderr: " + p->readAllStandardError(), LogError );
+        return false;
+    }
 
-void startGpgProcess(QProcess *p, const QStringList &args, QIODevice::OpenModeFlag mode)
-{
-    KeyPairPaths keys;
-    p->start(gpgExecutable(), getDefaultEncryptCommandArguments(keys.pub) + args, mode);
-}
-
-bool verifyProcess(QProcess *p)
-{
     const int exitCode = p->exitCode();
     if ( p->exitStatus() != QProcess::NormalExit ) {
         log( "ItemEncrypt ERROR: Failed to run GnuPG: " + p->errorString(), LogError );
@@ -101,16 +102,46 @@ bool verifyProcess(QProcess *p)
     return true;
 }
 
-bool waitOrTerminate(QProcess *p)
+bool checkGpgExecutable(const QString &executable)
 {
-    if ( p->state() != QProcess::NotRunning && !p->waitForFinished() ) {
-        p->terminate();
-        if ( !p->waitForFinished(5000) )
-            p->kill();
+    QProcess p;
+    p.start(executable, QStringList("--version"), QIODevice::ReadWrite);
+    p.closeReadChannel(QProcess::StandardError);
+
+    if ( !verifyProcess(&p, 5000) )
         return false;
+
+    const auto versionOutput = p.readAllStandardOutput();
+    return versionOutput.contains(" 2.");
+}
+
+QString findGpgExecutable()
+{
+    for (const auto &executable : {"gpg2", "gpg"}) {
+        if ( checkGpgExecutable(executable) )
+            return executable;
     }
 
-    return true;
+    return QString();
+}
+
+const QString &gpgExecutable()
+{
+    static const auto gpg = findGpgExecutable();
+    return gpg;
+}
+
+QStringList getDefaultEncryptCommandArguments(const QString &publicKeyPath)
+{
+    return QStringList() << "--trust-model" << "always" << "--recipient" << "copyq"
+                         << "--charset" << "utf-8" << "--display-charset" << "utf-8" << "--no-tty"
+                         << "--no-default-keyring" << "--keyring" << publicKeyPath;
+}
+
+void startGpgProcess(QProcess *p, const QStringList &args, QIODevice::OpenModeFlag mode)
+{
+    KeyPairPaths keys;
+    p->start(gpgExecutable(), getDefaultEncryptCommandArguments(keys.pub) + args, mode);
 }
 
 QString importGpgKey()
@@ -119,9 +150,6 @@ QString importGpgKey()
 
     QProcess p;
     p.start(gpgExecutable(), getDefaultEncryptCommandArguments(keys.pub) << "--import" << keys.sec);
-    if ( !waitOrTerminate(&p) )
-        return "Failed to import private key (process timed out).";
-
     if ( !verifyProcess(&p) )
         return "Failed to import private key (see log).";
 
@@ -138,9 +166,6 @@ QString exportGpgKey()
 
     QProcess p;
     p.start(gpgExecutable(), getDefaultEncryptCommandArguments(keys.pub) << "--export-secret-key" << "copyq");
-    if ( !waitOrTerminate(&p) )
-        return "Failed to export private key (process timed out).";
-
     if ( !verifyProcess(&p) )
         return "Failed to export private key (see log).";
 
@@ -233,16 +258,7 @@ QString exportImportGpgKeys()
 
 bool isGpgInstalled()
 {
-    QProcess p;
-    startGpgProcess(&p, QStringList("--version"), QIODevice::ReadOnly);
-    p.closeReadChannel(QProcess::StandardError);
-    p.waitForFinished();
-
-    if (p.exitStatus() != QProcess::NormalExit || p.exitCode() != 0)
-        return false;
-
-    const auto versionOutput = p.readAllStandardOutput();
-    return versionOutput.contains(" 2.");
+    return !gpgExecutable().isEmpty();
 }
 
 } // namespace
@@ -510,7 +526,7 @@ QString ItemEncryptedScriptable::generateTestKeys()
     QProcess process;
     startGenerateKeysProcess(&process, true);
 
-    if ( !waitOrTerminate(&process) || !verifyProcess(&process) ) {
+    if ( !verifyProcess(&process) ) {
         return QString("ItemEncrypt ERROR: %1; stderr: %2")
                 .arg( process.errorString(),
                       QString::fromUtf8(process.readAllStandardError()) );
@@ -690,7 +706,7 @@ ItemSaverPtr ItemEncryptedLoader::loadItems(const QString &, QAbstractItemModel
     // Wait for password entry dialog.
     p.waitForFinished(-1);
 
-    if ( !waitOrTerminate(&p) || !verifyProcess(&p) ) {
+    if ( !verifyProcess(&p) ) {
         emitDecryptFailed();
         return nullptr;
     }
