@@ -44,6 +44,8 @@ QString messageCodeToString(int code)
     switch (code) {
     case CommandFunctionCallReturnValue:
         return "CommandFunctionCallReturnValue";
+    case CommandInputDialogFinished:
+        return "CommandInputDialogFinished";
     case CommandStop:
         return "CommandStop";
     default:
@@ -122,9 +124,22 @@ void ClipboardClient::onMessageReceived(const QByteArray &data, int messageCode)
         emit functionCallResultReceived(data);
         break;
 
-    case CommandStop:
-        emit stopEventLoops();
+    case CommandInputDialogFinished: {
+        emit inputDialogFinished(data);
         break;
+    }
+
+    case CommandStop: {
+        emit stopEventLoops();
+
+        // Stop any event loops started later.
+        auto timer = new QTimer(this);
+        timer->setInterval(250);
+        connect(timer, &QTimer::timeout, this, &ClipboardClient::stopEventLoops);
+        connect(timer, &QTimer::timeout, timer, &QTimer::deleteLater);
+        timer->start();
+        break;
+    }
 
     default:
         log( "Unhandled message: " + messageCodeToString(messageCode), LogError );
@@ -165,21 +180,10 @@ void ClipboardClient::sendInput()
 
 void ClipboardClient::exit(int exitCode)
 {
-    emit functionCallResultReceived(QByteArray());
     abortInputReader();
     App::exit(exitCode);
 }
 
-void ClipboardClient::sendFunctionCall(const QByteArray &bytes)
-{
-    m_socket->sendMessage(bytes, CommandFunctionCall);
-
-    QEventLoop loop;
-    connect(this, &ClipboardClient::functionCallResultReceived, &loop, &QEventLoop::quit);
-    connect(qApp, &QCoreApplication::aboutToQuit, &loop, &QEventLoop::quit);
-    loop.exec();
-}
-
 void ClipboardClient::startInputReader()
 {
     if ( wasClosed() || m_inputReaderThread )
@@ -223,22 +227,31 @@ void ClipboardClient::start(const QStringList &arguments)
 
     connect( &scriptable, &Scriptable::readInput,
              this, &ClipboardClient::startInputReader );
-    connect( &scriptableProxy, &ScriptableProxy::sendFunctionCall,
-             this, &ClipboardClient::sendFunctionCall );
+    connect( &scriptableProxy, &ScriptableProxy::sendMessage,
+             m_socket, &ClientSocket::sendMessage );
 
     connect( this, &ClipboardClient::inputReceived,
              &scriptable, &Scriptable::setInput );
     connect( this, &ClipboardClient::functionCallResultReceived,
-             &scriptableProxy, &ScriptableProxy::setReturnValue );
+             &scriptableProxy, &ScriptableProxy::setFunctionCallReturnValue );
+    connect( this, &ClipboardClient::inputDialogFinished,
+             &scriptableProxy, &ScriptableProxy::setInputDialogResult );
 
     connect( m_socket, &ClientSocket::disconnected,
              &scriptable, &Scriptable::abort );
+    connect( m_socket, &ClientSocket::disconnected,
+             &scriptableProxy, &ScriptableProxy::clientDisconnected );
 
     connect( this, &ClipboardClient::stopEventLoops,
              &scriptable, &Scriptable::stopEventLoops );
     connect( qApp, &QCoreApplication::aboutToQuit,
              &scriptable, &Scriptable::stopEventLoops );
 
+    connect( &scriptable, &Scriptable::finished,
+             this, &ClipboardClient::scriptableFinished );
+    connect( &scriptable, &Scriptable::finished,
+             &scriptableProxy, &ScriptableProxy::clientDisconnected );
+
     bool hasData;
     auto actionId = qgetenv("COPYQ_ACTION_ID").toInt(&hasData);
     if (!hasData)
