@@ -173,7 +173,7 @@ void ClipboardServer::stopMonitoring()
 
 void ClipboardServer::startMonitoring()
 {
-    if (m_monitor || m_preventMonitorStart || !m_wnd->isMonitoringEnabled())
+    if (m_monitor || m_ignoreNewConnections || !m_wnd->isMonitoringEnabled())
         return;
 
     COPYQ_LOG("Starting monitor");
@@ -226,7 +226,7 @@ void ClipboardServer::onAboutToQuit()
 {
     COPYQ_LOG("Closing server.");
 
-    m_preventMonitorStart = true;
+    m_ignoreNewConnections = true;
 
     terminateClients(10000);
     m_server->close(); // No new connections can be open.
@@ -239,13 +239,13 @@ void ClipboardServer::onCommitData(QSessionManager &sessionManager)
 {
     COPYQ_LOG("Got commit data request from session manager.");
 
-    m_preventMonitorStart = true;
+    m_ignoreNewConnections = true;
     const bool cancel = sessionManager.allowsInteraction() && !askToQuit();
     sessionManager.release();
 
     if (cancel) {
         sessionManager.cancel();
-        m_preventMonitorStart = false;
+        m_ignoreNewConnections = false;
         startMonitoring();
     } else {
         m_wnd->saveTabs();
@@ -344,7 +344,18 @@ void ClipboardServer::waitForClientsToFinish(int waitMs)
 
 void ClipboardServer::onClientNewConnection(const ClientSocketPtr &client)
 {
-    auto proxy = new ScriptableProxy(m_wnd, client.get());
+    if (m_ignoreNewConnections) {
+        COPYQ_LOG("Ignoring new client while exiting");
+        client->close();
+        return;
+    }
+
+    auto proxy = new ScriptableProxy(m_wnd);
+    connect( client.get(), &ClientSocket::destroyed,
+             proxy, &ScriptableProxy::deleteLater );
+    connect( proxy, &ScriptableProxy::sendMessage,
+             client.get(), &ClientSocket::sendMessage );
+
     m_clients.insert( client->id(), ClientData(client, proxy) );
     connect( this, &ClipboardServer::closeClients,
              client.get(), &ClientSocket::close );
@@ -352,6 +363,8 @@ void ClipboardServer::onClientNewConnection(const ClientSocketPtr &client)
              this, &ClipboardServer::onClientMessageReceived );
     connect( client.get(), &ClientSocket::disconnected,
              this, &ClipboardServer::onClientDisconnected );
+    connect( client.get(), &ClientSocket::disconnected,
+             proxy, &ScriptableProxy::clientDisconnected );
     connect( client.get(), &ClientSocket::connectionFailed,
              this, &ClipboardServer::onClientConnectionFailed );
     client->start();
@@ -366,8 +379,7 @@ void ClipboardServer::onClientMessageReceived(
         if (!clientData.isValid())
             return;
 
-        const auto result = clientData.proxy->callFunction(message);
-        clientData.client->sendMessage(result, CommandFunctionCallReturnValue);
+        clientData.proxy->callFunction(message);
         break;
     }
     default:
