@@ -22,7 +22,6 @@
 #include "common/common.h"
 #include "common/log.h"
 #include "common/mimetypes.h"
-#include "common/textdata.h"
 #include "item/serialize.h"
 
 #include <QCoreApplication>
@@ -181,7 +180,7 @@ Action::~Action()
     closeSubCommands();
 }
 
-QString Action::command() const
+QString Action::commandLine() const
 {
     QString text;
     for ( const auto &line : m_cmds ) {
@@ -206,7 +205,7 @@ void Action::setCommand(const QStringList &arguments)
     m_cmds.append(QList<QStringList>() << arguments);
 }
 
-void Action::setInput(const QVariantMap &data, const QString &inputFormat)
+void Action::setInputWithFormat(const QVariantMap &data, const QString &inputFormat)
 {
     if (inputFormat == mimeItems) {
         m_input = serializeData(data);
@@ -319,7 +318,6 @@ bool Action::isRunning() const
 void Action::setData(const QVariantMap &data)
 {
     m_data = data;
-    emit dataChanged(data);
 }
 
 const QVariantMap &Action::data() const
@@ -327,6 +325,17 @@ const QVariantMap &Action::data() const
     return m_data;
 }
 
+void Action::appendOutput(const QByteArray &output)
+{
+    if ( !output.isEmpty() )
+        emit actionOutput(output);
+}
+
+void Action::appendErrorOutput(const QByteArray &errorOutput)
+{
+    m_errorOutput.append(errorOutput);
+}
+
 void Action::onSubProcessError(QProcess::ProcessError error)
 {
     QProcess *p = qobject_cast<QProcess*>(sender());
@@ -362,14 +371,8 @@ void Action::onSubProcessOutput()
         return;
 
     auto p = m_processes.last();
-    if ( !p->isReadable() )
-        return;
-
-    const auto output = p->readAll();
-    if ( output.isEmpty() )
-        return;
-
-    emit actionOutput(output);
+    if ( p->isReadable() )
+        appendOutput( p->readAll() );
 }
 
 void Action::onSubProcessErrorOutput()
@@ -378,7 +381,7 @@ void Action::onSubProcessErrorOutput()
     Q_ASSERT(p);
 
     if ( p->isReadable() )
-        m_errorOutput.append( getTextData(p->readAllStandardError()) );
+        appendErrorOutput( p->readAllStandardError() );
 }
 
 void Action::writeInput()
