@@ -690,6 +690,13 @@ void ClipboardBrowser::dragDropScroll()
     }
 }
 
+void ClipboardBrowser::setCurrentIndex(const QModelIndex &index)
+{
+    // WORKAROUND: QAbstractItemView::setCurrentIndex() seems to depend on
+    //             currently pressed keyboard modifiers, which is unexpected.
+    selectionModel()->setCurrentIndex(index, QItemSelectionModel::ClearAndSelect);
+}
+
 QVariantMap ClipboardBrowser::copyIndex(const QModelIndex &index) const
 {
     auto data = index.data(contentType::data).toMap();
@@ -1323,15 +1330,23 @@ void ClipboardBrowser::filterItems(const QRegExp &re)
         m_filterRow = -1;
 
     int row = 0;
-    for ( ; row < length() && hideFiltered(row); ++row ) {}
 
-    setCurrentIndex(index(row));
+    if ( re.isEmpty() ) {
+        for ( ; row < length(); ++row )
+            hideFiltered(row);
 
-    for ( ; row < length(); ++row )
-        hideFiltered(row);
+        scrollTo(currentIndex(), PositionAtCenter);
+    } else {
+        for ( ; row < length() && hideFiltered(row); ++row ) {}
 
-    if ( filterByRowNumber && m_filterRow >= 0 && m_filterRow < m.rowCount() )
-        setCurrentIndex( index(m_filterRow) );
+        setCurrent(row);
+
+        for ( ; row < length(); ++row )
+            hideFiltered(row);
+
+        if ( filterByRowNumber && m_filterRow >= 0 && m_filterRow < m.rowCount() )
+            setCurrent(m_filterRow);
+    }
 }
 
 void ClipboardBrowser::moveToClipboard(const QModelIndex &ind)
@@ -1519,8 +1534,7 @@ void ClipboardBrowser::setCurrent(int row, bool keepSelection, bool setCurrentOn
         else if (!currentSelected)
             sel->setCurrentIndex(prev, QItemSelectionModel::Deselect);
     } else {
-        const auto ind = index(i);
-        selectionModel()->setCurrentIndex(ind, QItemSelectionModel::ClearAndSelect);
+        setCurrentIndex( index(i) );
     }
 }
 
@@ -1607,14 +1621,13 @@ bool ClipboardBrowser::add(const QVariantMap &data, int row)
     return true;
 }
 
-void ClipboardBrowser::addUnique(const QVariantMap &data)
+void ClipboardBrowser::addUnique(const QVariantMap &data, ClipboardMode mode)
 {
     if ( moveToTop(hash(data)) ) {
         COPYQ_LOG("New item: Moving existing to top");
         return;
     }
 
-#ifdef HAS_MOUSE_SELECTIONS
     // When selecting text under X11, clipboard data may change whenever selection changes.
     // Instead of adding item for each selection change, this updates previously added item.
     // Also update previous item if the same selected text is copied to clipboard afterwards.
@@ -1629,8 +1642,7 @@ void ClipboardBrowser::addUnique(const QVariantMap &data)
         {
             const auto newText = getTextData(data);
             const auto oldText = getTextData(previousData);
-            const auto isClipboard = isClipboardData(data);
-            if ( isClipboard
+            if ( (mode == ClipboardMode::Clipboard)
                  ? (newText == oldText)
                  : getTextData(data).contains(getTextData(previousData)) )
             {
@@ -1648,7 +1660,6 @@ void ClipboardBrowser::addUnique(const QVariantMap &data)
             }
         }
     }
-#endif
 
     COPYQ_LOG("New item: Adding");
 
