@@ -194,8 +194,6 @@ public:
     bool setCurrentTab(int index);
     /** Show window and given tab and give focus to the tab. */
     void showBrowser(int index);
-    /** Enter browse mode and reset search. */
-    void resetStatus();
 
     /** Open tab group renaming dialog. */
     void renameTabGroup(const QString &name);
@@ -280,6 +278,8 @@ public:
     QVector<Command> displayCommands() const { return m_displayCommands; }
     QVector<Command> scriptCommands() const { return m_scriptCommands; }
 
+    void snip();
+
 public slots:
     /** Close main window and exit the application. */
     void exit();
@@ -404,34 +404,10 @@ public slots:
 
     /** Set text for filtering items. */
     void setFilter(const QString &text);
-
-    void hideSearchBar();
+    QString filter() const;
 
     void updateShortcuts();
 
-#ifdef HAS_TESTS
-    /**
-     * Send key clicks to currently focused widget.
-     *
-     * This function will block if it opens a modal dialog.
-     * So use sendKeyClicks() instead.
-     *
-     * Increments key clicks sequence number returned by lastReceivedKeyClicks().
-     */
-    void keyClicks(const QString &keys, int delay);
-
-    /**
-     * Send key clicks to focused widget.
-     * @return Key clicks sequence number.
-     */
-    uint sendKeyClicks(const QString &keys, int delay);
-
-    /**
-     * @return Last key clicks sequence number received by widgets.
-     */
-    uint lastReceivedKeyClicks();
-#endif
-
 signals:
     /** Request clipboard change. */
     void changeClipboard(const QVariantMap &data, ClipboardMode mode);
@@ -454,8 +430,6 @@ protected:
     /** Hide (minimize to tray) window on close. */
     void closeEvent(QCloseEvent *event) override;
 
-    void showEvent(QShowEvent *event) override;
-
     bool focusNextPrevChild(bool next) override;
 
     bool nativeEvent(const QByteArray &eventType, void *message, long *result) override;
@@ -640,8 +614,8 @@ private:
 
     void addMenuItems(TrayMenu *menu, ClipboardBrowser *c, int maxItemCount, const QString &searchText);
     void activateMenuItem(ClipboardBrowser *c, const QVariantMap &data, bool omitPaste);
-    QWidget *toggleMenu(TrayMenu *menu, QPoint pos);
-    QWidget *toggleMenu(TrayMenu *menu);
+    bool toggleMenu(TrayMenu *menu, QPoint pos);
+    bool toggleMenu(TrayMenu *menu);
 
     bool exportData(const QString &fileName, const QStringList &tabs, bool exportConfiguration, bool exportCommands);
     bool exportDataV3(QDataStream *out, const QStringList &tabs, bool exportConfiguration, bool exportCommands);
@@ -679,7 +653,6 @@ private:
     QTimer m_timerUpdateFocusWindows;
     QTimer m_timerUpdateContextMenu;
     QTimer m_timerUpdateTrayMenu;
-    QTimer m_timerShowWindow;
     QTimer m_timerTrayAvailable;
     QTimer m_timerTrayIconSnip;
     QTimer m_timerSaveTabPositions;
@@ -698,7 +671,8 @@ private:
 
     QPointer<CommandDialog> m_commandDialog;
 
-    bool m_iconSnip;
+    bool m_iconSnip = false;
+    bool m_forceIconSnip = false;
 
     bool m_wasMaximized;
 
@@ -720,12 +694,7 @@ private:
 
     ClipboardManager m_clipboardManager;
 
-#ifdef HAS_TESTS
-    /// Key clicks sequence number last returned by sendKeyClicks().
-    uint m_sentKeyClicks = 0;
-    /// Key clicks sequence number received by widgets.
-    uint m_receivedKeyClicks = 0;
-#endif
+    bool m_isActiveWindow = false;
 };
 
 #endif // MAINWINDOW_H
