@@ -56,6 +56,7 @@
 #include <QScriptEngine>
 #include <QScriptValueIterator>
 #include <QSettings>
+#include <QSysInfo>
 #include <QUrl>
 #include <QVector>
 #include <QTextCodec>
@@ -738,10 +739,26 @@ QScriptValue Scriptable::call(QScriptValue *fn, const QScriptValue &object, cons
 QScriptValue Scriptable::version()
 {
     m_skipArguments = 0;
-    return tr(programName) + " " COPYQ_VERSION " (hluk@email.cz)\n"
-            + tr("Built with: ")
-            + "Qt " + QT_VERSION_STR +
-            + '\n';
+    return tr(programName) + " " COPYQ_VERSION "\n"
+            + "Qt: " QT_VERSION_STR "\n"
+            + "Compiler: "
+#if defined(Q_CC_GNU)
+            "GCC"
+#elif defined(Q_CC_CLANG)
+            "Clang"
+#elif defined(Q_CC_MINGW)
+            "MinGW"
+#elif defined(Q_CC_MSVC)
+            "MSVC"
+#else
+            "???"
+#endif
+            + "\n"
+#if QT_VERSION >= QT_VERSION_CHECK(5,4,0)
+            + "Arch: " + QSysInfo::buildAbi() + "\n"
+            + "OS: " + QSysInfo::prettyProductName() + "\n"
+#endif
+            ;
 }
 
 QScriptValue Scriptable::help()
@@ -757,7 +774,7 @@ QScriptValue Scriptable::help()
             helpString.append(hlp.toString());
 
         helpString.append("\n" + helpTail() + "\n\n" + tr(programName)
-            + " " + COPYQ_VERSION + " (hluk@email.cz)\n");
+            + " " + COPYQ_VERSION + "\n");
     } else {
         for (int i = 0; i < argumentCount(); ++i) {
             const QString &cmd = toString(argument(i), this);
@@ -903,10 +920,14 @@ QScriptValue Scriptable::focused()
     return m_proxy->isMainWindowFocused();
 }
 
-void Scriptable::filter()
+QScriptValue Scriptable::filter()
 {
     m_skipArguments = 1;
+    if (argumentCount() == 0)
+        return m_proxy->filter();
+
     m_proxy->filter(arg(0));
+    return QScriptValue();
 }
 
 void Scriptable::ignore()
@@ -953,7 +974,7 @@ QScriptValue Scriptable::hasSelectionFormat()
 
 QScriptValue Scriptable::isClipboard()
 {
-    return !m_data.keys().contains(mimeClipboardMode);
+    return isClipboardData(m_data);
 }
 
 QScriptValue Scriptable::copy()
@@ -986,7 +1007,7 @@ QScriptValue Scriptable::tab()
     if ( name.isNull() )
         return toScriptValue( m_proxy->tabs(), this );
 
-    m_proxy->setTab(name);
+    m_tabName = name;
     return QScriptValue();
 }
 
@@ -1028,7 +1049,7 @@ QScriptValue Scriptable::tabIcon()
 QScriptValue Scriptable::length()
 {
     m_skipArguments = 0;
-    return m_proxy->browserLength();
+    return m_proxy->browserLength(m_tabName);
 }
 
 void Scriptable::select()
@@ -1039,7 +1060,7 @@ void Scriptable::select()
     if ( !toInt(value, &row) )
         throwError(argumentError());
 
-    m_proxy->browserMoveToClipboard(row);
+    m_proxy->browserMoveToClipboard(m_tabName, row);
 }
 
 void Scriptable::next()
@@ -1072,7 +1093,7 @@ void Scriptable::remove()
     if ( rows.empty() )
         rows.append(0);
 
-    const auto error = m_proxy->browserRemoveRows(rows);
+    const auto error = m_proxy->browserRemoveRows(m_tabName, rows);
     if ( !error.isEmpty() )
         throwError(error);
 }
@@ -1091,7 +1112,7 @@ void Scriptable::edit()
         if (i > 0)
             text.append(m_inputSeparator);
         if ( toInt(value, &row) ) {
-            const QByteArray bytes = row >= 0 ? m_proxy->browserItemData(row, mimeText)
+            const QByteArray bytes = row >= 0 ? m_proxy->browserItemData(m_tabName, row, mimeText)
                                               : m_proxy->getClipboardData(mimeText);
             text.append( getTextData(bytes) );
         } else {
@@ -1101,13 +1122,13 @@ void Scriptable::edit()
 
     bool changeClipboard = row < 0;
 
-    if ( !m_proxy->browserOpenEditor(fromString(text), changeClipboard) ) {
-        m_proxy->showCurrentBrowser();
+    if ( !m_proxy->browserOpenEditor(m_tabName, fromString(text), changeClipboard) ) {
+        m_proxy->showBrowser(m_tabName);
         if (len == 1 && row >= 0) {
-            m_proxy->browserSetCurrent(row);
-            m_proxy->browserEditRow(row);
+            m_proxy->browserSetCurrent(m_tabName, row);
+            m_proxy->browserEditRow(m_tabName, row);
         } else {
-            m_proxy->browserEditNew(text, changeClipboard);
+            m_proxy->browserEditNew(m_tabName, text, changeClipboard);
         }
     }
 }
@@ -1128,7 +1149,7 @@ QScriptValue Scriptable::read()
             if (used)
                 result.append( m_inputSeparator.toUtf8() );
             used = true;
-            result.append( row >= 0 ? m_proxy->browserItemData(row, mime)
+            result.append( row >= 0 ? m_proxy->browserItemData(m_tabName, row, mime)
                                     : m_proxy->getClipboardData(mime) );
         } else {
             mime = toString(value, this);
@@ -1175,7 +1196,7 @@ void Scriptable::action()
             text.append(m_inputSeparator);
         else
             anyRows = true;
-        text.append( getTextData(m_proxy->browserItemData(row, mimeText)) );
+        text.append( getTextData(m_proxy->browserItemData(m_tabName, row, mimeText)) );
     }
 
     m_skipArguments = i + 2;
@@ -1192,7 +1213,7 @@ void Scriptable::action()
         command.output = mimeText;
         command.input = mimeText;
         command.wait = false;
-        command.outputTab = m_proxy->tab();
+        command.outputTab = m_proxy->tab(m_tabName);
         command.sep = ((i + 1) < argumentCount()) ? toString( argument(i + 1), this )
                                                   : QString('\n');
         m_proxy->action(data, command);
@@ -1262,7 +1283,7 @@ void Scriptable::exportTab()
     const auto filePath = arg(0);
     if ( filePath.isNull() )
         throwError(argumentError());
-    else if ( !m_proxy->saveTab(getAbsoluteFilePath(filePath)) )
+    else if ( !m_proxy->saveTab(m_tabName, getAbsoluteFilePath(filePath)) )
         throwSaveError(filePath);
 }
 
@@ -1628,14 +1649,30 @@ void Scriptable::keys()
     if (!ok)
         delay = 0;
 
+    QString expectedWidgetName;
+
+    const auto focusPrefix = QLatin1String("focus:");
     for (int i = 0; i < argumentCount(); ++i) {
         const QString keys = toString(argument(i), this);
 
-        waitFor(wait);
-        m_proxy->sendKeys(keys, delay);
+        if (keys.startsWith(focusPrefix)) {
+            expectedWidgetName = keys.mid(focusPrefix.size());
+            m_proxy->sendKeys(expectedWidgetName, QString(), 0);
+        } else {
+            waitFor(wait);
+            m_proxy->sendKeys(expectedWidgetName, keys, delay);
+        }
 
         // Make sure all keys are send (shortcuts are postponed because they can be blocked by modal windows).
-        while ( !m_proxy->keysSent() ) {
+        for (;;) {
+            if ( m_proxy->sendKeysSucceeded() )
+                break;
+
+            if ( m_proxy->sendKeysFailed() ) {
+                throwError("Failed to send key presses");
+                return;
+            }
+
             QCoreApplication::processEvents();
             if (!canContinue()) {
                 throwError("Disconnected");
@@ -1680,7 +1717,7 @@ QScriptValue Scriptable::selectItems()
 {
     const auto rows = getRows();
     m_skipArguments = rows.size();
-    return m_proxy->selectItems(rows);
+    return m_proxy->selectItems(m_tabName, rows);
 }
 
 QScriptValue Scriptable::selectedTab()
@@ -1773,7 +1810,7 @@ QScriptValue Scriptable::getItem()
         return QScriptValue();
     }
 
-    return toScriptValue( m_proxy->browserItemData(row), this );
+    return toScriptValue( m_proxy->browserItemData(m_tabName, row), this );
 }
 
 void Scriptable::setItem()
@@ -1875,7 +1912,21 @@ QScriptValue Scriptable::dialog()
         values.append( NamedValue(key, toVariant(value)) );
     }
 
-    values = m_proxy->inputDialog(values);
+    const auto dialogId = m_proxy->inputDialog(values);
+    if ( !canContinue() )
+        return QScriptValue();
+
+    QEventLoop loop;
+    connect(this, &Scriptable::finished, &loop, &QEventLoop::quit);
+    connect(this, &Scriptable::stop, &loop, &QEventLoop::quit);
+    connect( m_proxy, &ScriptableProxy::inputDialogFinished,
+             &loop, [&](int finishedDialogId, const NamedValueList &result) {
+                 if (finishedDialogId != dialogId)
+                     return;
+                 values = result;
+                 loop.quit();
+             });
+    loop.exec();
 
     if (values.isEmpty())
         return QScriptValue();
@@ -2214,10 +2265,8 @@ void Scriptable::saveData()
 {
     const QString outputTab = getTextData(m_data, mimeOutputTab);
 
-    if ( !outputTab.isEmpty() ) {
-        auto data = copyWithoutInternalData(m_data);
-        m_proxy->saveData(outputTab, data);
-    }
+    if ( !outputTab.isEmpty() )
+        saveData(outputTab);
 }
 
 QScriptValue Scriptable::hasData()
@@ -2345,12 +2394,13 @@ void Scriptable::onMonitorClipboardChanged(const QVariantMap &data, ClipboardOwn
     m_data = data;
     setActionData();
 
-    if (ownership == ClipboardOwnership::Own)
-        onOwnClipboardChanged();
-    else if (ownership == ClipboardOwnership::Hidden)
-        onHiddenClipboardChanged();
-    else
-        onClipboardChanged();
+    const auto functionName =
+        ownership == ClipboardOwnership::Own ? "onOwnClipboardChanged"
+      : ownership == ClipboardOwnership::Hidden ? "onHiddenClipboardChanged"
+      : "onClipboardChanged";
+
+    auto function = m_engine->globalObject().property(functionName);
+    function.call();
 
     processUncaughtMonitorException("ClipboardMonitor::onMonitorClipboardChanged");
 }
@@ -2648,17 +2698,17 @@ void Scriptable::changeItem(bool create)
 
     if (create) {
         QVector<QVariantMap> items(1, data);
-        const auto error = m_proxy->browserInsert(row, items);
+        const auto error = m_proxy->browserInsert(m_tabName, row, items);
         if ( !error.isEmpty() )
             throwError(error);
     } else {
-        m_proxy->browserChange(data, row);
+        m_proxy->browserChange(m_tabName, data, row);
     }
 }
 
 void Scriptable::nextToClipboard(int where)
 {
-    QVariantMap data = m_proxy->nextItem(where);
+    QVariantMap data = m_proxy->nextItem(m_tabName, where);
     if (data.isEmpty())
         return;
 
@@ -2847,10 +2897,8 @@ bool Scriptable::runCommands(CommandType::CommandType type)
         }
 
         if ( type == CommandType::Automatic ) {
-            if ( !command.tab.isEmpty() ) {
-                auto data = copyWithoutInternalData(m_data);
-                m_proxy->saveData(command.tab, data);
-            }
+            if ( !command.tab.isEmpty() )
+                saveData(command.tab);
 
             if ( command.remove || command.transform || m_data.contains(mimeIgnore) ) {
                 COPYQ_LOG( QString(label).arg(command.name, "Ignoring data") );
@@ -2976,7 +3024,7 @@ void Scriptable::insert(int row, int argumentsBegin, int argumentsEnd)
             items.append( createDataMap(mimeText, toString(arg, this)) );
     }
 
-    const auto error = m_proxy->browserInsert(row, items);
+    const auto error = m_proxy->browserInsert(m_tabName, row, items);
     if ( !error.isEmpty() )
         throwError(error);
 }
@@ -3053,7 +3101,7 @@ void Scriptable::synchronizeSelection(ClipboardMode targetMode)
         const auto source = targetMode == ClipboardMode::Selection ? QClipboard::Clipboard : QClipboard::Selection;
 
         // Stop if the clipboard/selection text already changed again.
-        auto clipboard = QGuiApplication::clipboard();
+        auto clipboard = systemClipboard();
         const auto sourceText = clipboard->text(source);
         if (sourceText != getTextData(m_data)) {
             COPYQ_SYNC_LOG("Cancelled (source text changed)");
@@ -3087,6 +3135,15 @@ void Scriptable::synchronizeSelection(ClipboardMode targetMode)
 #endif
 }
 
+void Scriptable::saveData(const QString &tab)
+{
+    const auto data = copyWithoutInternalData(m_data);
+    const auto clipboardMode = isClipboardData(m_data)
+            ? ClipboardMode::Clipboard
+            : ClipboardMode::Selection;
+    m_proxy->saveData(tab, data, clipboardMode);
+}
+
 QScriptValue NetworkReply::get(const QString &url, Scriptable *scriptable)
 {
     NetworkReply *reply = new NetworkReply(url, QByteArray(), scriptable);
