@@ -22,6 +22,7 @@
 #include "common/action.h"
 #include "common/appconfig.h"
 #include "common/command.h"
+#include "common/commandstatus.h"
 #include "common/commandstore.h"
 #include "common/common.h"
 #include "common/config.h"
@@ -77,37 +78,54 @@
 #include <QTextEdit>
 #include <QUrl>
 
+#ifdef HAS_TESTS
+#   include <QTest>
+#endif
+
 #include <type_traits>
 
 const quint32 serializedFunctionCallMagicNumber = 0x58746908;
-const quint32 serializedFunctionCallVersion = 1;
+const quint32 serializedFunctionCallVersion = 2;
 
-#define BROWSER(call) \
-    ClipboardBrowser *c = fetchBrowser(); \
+#define BROWSER(tabName, call) \
+    ClipboardBrowser *c = fetchBrowser(tabName); \
     if (c) \
         (c->call)
 
 #define STR(str) str
 
-#define INVOKE_(function, arguments) \
+#define INVOKE_(function, arguments, functionCallId) \
     static auto f = FunctionCallSerializer(STR(#function)).withSlotArguments arguments; \
-    f.setTabName(m_tabName); \
     f.setArguments arguments; \
-    emit sendFunctionCall(f.serializeAndClear()) \
+    emit sendMessage(f.serializeAndClear(functionCallId), CommandFunctionCall)
 
-#define INVOKE(function, arguments) \
+#define INVOKE_NO_SNIP(function, arguments) \
     if (!m_wnd) { \
         using Result = decltype(function arguments); \
-        INVOKE_(function, arguments); \
-        return m_returnValue.value<Result>(); \
+        const auto functionCallId = ++m_lastFunctionCallId; \
+        INVOKE_(function, arguments, functionCallId); \
+        const auto result = waitForFunctionCallFinished(functionCallId); \
+        return result.value<Result>(); \
     }
 
-#define INVOKE2(function, arguments) \
+#define INVOKE_NO_SNIP2(FUNCTION, ARGUMENTS) \
     if (!m_wnd) { \
-        INVOKE_(function, arguments); \
+        const auto functionCallId = ++m_lastFunctionCallId; \
+        INVOKE_(FUNCTION, ARGUMENTS, functionCallId); \
+        waitForFunctionCallFinished(functionCallId); \
         return; \
     }
 
+#define INVOKE(FUNCTION, ARGUMENTS) \
+    INVOKE_NO_SNIP(FUNCTION, ARGUMENTS) \
+    if (m_wnd) \
+        m_wnd->snip()
+
+#define INVOKE2(FUNCTION, ARGUMENTS) \
+    INVOKE_NO_SNIP2(FUNCTION, ARGUMENTS) \
+    if (m_wnd) \
+        m_wnd->snip()
+
 Q_DECLARE_METATYPE(QFile*)
 
 QDataStream &operator<<(QDataStream &out, const NotificationButton &button)
@@ -255,13 +273,8 @@ const char propertyWidgetName[] = "CopyQ_widget_name";
 const char propertyWidgetProperty[] = "CopyQ_widget_property";
 
 struct InputDialog {
-    QDialog dialog;
+    QDialog *dialog = nullptr;
     QString defaultChoice; /// Default text for list widgets.
-
-    explicit InputDialog(QWidget *parentWindow)
-        : dialog(parentWindow)
-    {
-    }
 };
 
 template<typename ...Ts>
@@ -290,7 +303,7 @@ class FunctionCallSerializer {
 public:
     explicit FunctionCallSerializer(const char *functionName)
     {
-        m_args << QVariant() << QByteArray(functionName);
+        m_slotName = functionName;
     }
 
     template<typename ...Ts>
@@ -302,13 +315,14 @@ public:
         return *this;
     }
 
-    QByteArray serializeAndClear()
+    QByteArray serializeAndClear(int functionCallId)
     {
         QByteArray bytes;
         QDataStream stream(&bytes, QIODevice::WriteOnly);
         stream.setVersion(QDataStream::Qt_5_0);
-        stream << serializedFunctionCallMagicNumber << serializedFunctionCallVersion << m_args;
-        m_args.resize(2);
+        stream << serializedFunctionCallMagicNumber << serializedFunctionCallVersion
+               << functionCallId << m_slotName << m_args;
+        m_args.clear();
         return bytes;
     }
 
@@ -321,24 +335,19 @@ public:
         setArguments(args...);
     }
 
-    void setTabName(const QString &tabName)
-    {
-        m_args[0] = tabName;
-    }
-
 private:
     void setSlotArgumentTypes(const QByteArray &args)
     {
-        const auto slotName = m_args[1].toByteArray() + "(" + args + ")";
-        const int slotIndex = ScriptableProxy::staticMetaObject.indexOfSlot(slotName);
+        m_slotName += "(" + args + ")";
+        const int slotIndex = ScriptableProxy::staticMetaObject.indexOfSlot(m_slotName);
         if (slotIndex == -1) {
-            log("Failed to find scriptable proxy slot: " + slotName, LogError);
+            log("Failed to find scriptable proxy slot: " + m_slotName, LogError);
             Q_ASSERT(false);
         }
-        m_args[1] = slotName;
     }
 
     QVector<QVariant> m_args;
+    QByteArray m_slotName;
 };
 
 class ScreenshotRectWidget : public QLabel {
@@ -476,7 +485,7 @@ void installShortcutToCloseDialog(QWidget *dialog, QWidget *shortcutParent, int
 
 QWidget *createListWidget(const QString &name, const QStringList &items, InputDialog *inputDialog)
 {
-    QDialog *parent = &inputDialog->dialog;
+    QDialog *parent = inputDialog->dialog;
 
     const QString currentText = inputDialog->defaultChoice.isNull()
             ? items.value(0)
@@ -557,7 +566,7 @@ QWidget *createTextEdit(const QString &name, const QVariant &value, QWidget *par
 
 QWidget *createWidget(const QString &name, const QVariant &value, InputDialog *inputDialog)
 {
-    QDialog *parent = &inputDialog->dialog;
+    QDialog *parent = inputDialog->dialog;
 
     switch ( value.type() ) {
     case QVariant::Bool:
@@ -613,7 +622,7 @@ QString tabNameEmptyError()
     return ScriptableProxy::tr("Tab name cannot be empty!");
 }
 
-void raiseWindow(QWidget *window)
+void raiseWindow(QPointer<QWidget> window)
 {
     window->raise();
     window->activateWindow();
@@ -627,10 +636,184 @@ void raiseWindow(QWidget *window)
 
 } // namespace
 
+#ifdef HAS_TESTS
+class KeyClicker : public QObject {
+public:
+    KeyClicker(MainWindow *wnd, QObject *parent)
+        : QObject(parent)
+        , m_wnd(wnd)
+    {
+    }
+
+    void keyClicksRetry(const QString &expectedWidgetName, const QString &keys, int delay, int retry)
+    {
+        if (retry > 0)
+            sendKeyClicks(expectedWidgetName, keys, delay + 100, retry - 1);
+        else
+            keyClicksFailed(expectedWidgetName);
+    }
+
+    void keyClicksFailed(const QString &expectedWidgetName)
+    {
+        auto actual = keyClicksTarget();
+        auto popup = QApplication::activePopupWidget();
+        auto widget = QApplication::focusWidget();
+        auto window = QApplication::activeWindow();
+        auto modal = QApplication::activeModalWidget();
+        const auto currentWindow = createPlatformNativeInterface()->getCurrentWindow();
+        const auto currentWindowTitle = currentWindow ? currentWindow->getTitle() : QString();
+        log( QString("Failed to send key press to target widget")
+            + "\nExpected: " + (expectedWidgetName.isEmpty() ? "Any" : expectedWidgetName)
+            + "\nActual:   " + keyClicksTargetDescription(actual)
+            + "\nPopup:    " + keyClicksTargetDescription(popup)
+            + "\nWidget:   " + keyClicksTargetDescription(widget)
+            + "\nWindow:   " + keyClicksTargetDescription(window)
+            + "\nModal:    " + keyClicksTargetDescription(modal)
+            + "\nTitle:    " + currentWindowTitle
+            , LogError );
+
+        m_failed = true;
+    }
+
+    void keyClicks(const QString &expectedWidgetName, const QString &keys, int delay, int retry)
+    {
+        auto widget = keyClicksTarget();
+        if (!widget) {
+            keyClicksRetry(expectedWidgetName, keys, delay, retry);
+            return;
+        }
+
+        auto widgetName = keyClicksTargetDescription(widget);
+        if ( !expectedWidgetName.isEmpty() && !widgetName.contains(expectedWidgetName) ) {
+            keyClicksRetry(expectedWidgetName, keys, delay, retry);
+            return;
+        }
+
+        // Only verified focused widget.
+        if ( keys.isEmpty() ) {
+            m_succeeded = true;
+            return;
+        }
+
+        // There could be some animation/transition effect on check boxes
+        // so wait for checkbox to be set.
+        if ( qobject_cast<QCheckBox*>(widget) )
+            waitFor(100);
+
+        COPYQ_LOG( QString("Sending keys \"%1\" to %2.")
+                   .arg(keys, widgetName) );
+
+        const auto popupMessage = QString("%1 (%2)")
+                .arg( quoteString(keys), widgetName );
+        auto notification = m_wnd->createNotification();
+        notification->setMessage(popupMessage);
+        notification->setIcon(IconKeyboard);
+        notification->setInterval(2000);
+
+        if ( keys.startsWith(":") ) {
+            const auto text = keys.mid(1);
+
+            QTest::keyClicks(widget, text, Qt::NoModifier, 0);
+
+            // Increment key clicks sequence number after typing all the text.
+            m_succeeded = true;
+        } else {
+            const QKeySequence shortcut(keys, QKeySequence::PortableText);
+
+            if ( shortcut.isEmpty() ) {
+                log( QString("Cannot parse shortcut \"%1\"!").arg(keys), LogError );
+                m_failed = true;
+                return;
+            }
+
+            // Increment key clicks sequence number before opening any modal dialogs.
+            m_succeeded = true;
+
+            const auto key = static_cast<uint>(shortcut[0]);
+            QTest::keyClick( widget,
+                             Qt::Key(key & ~Qt::KeyboardModifierMask),
+                             Qt::KeyboardModifiers(key & Qt::KeyboardModifierMask),
+                             0 );
+        }
+
+        COPYQ_LOG( QString("Key \"%1\" sent to %2.")
+                   .arg(keys, widgetName) );
+    }
+
+    void sendKeyClicks(const QString &expectedWidgetName, const QString &keys, int delay, int retry)
+    {
+        m_succeeded = false;
+        m_failed = false;
+
+        // Don't stop when modal window is open.
+        auto t = new QTimer(m_wnd);
+        t->setSingleShot(true);
+        QObject::connect( t, &QTimer::timeout, this, [=]() {
+            keyClicks(expectedWidgetName, keys, delay, retry);
+            t->deleteLater();
+        });
+        t->start(delay);
+    }
+
+    bool succeeded() const { return m_succeeded; }
+    bool failed() const { return m_failed; }
+
+private:
+    static QString keyClicksTargetDescription(QWidget *widget)
+    {
+        if (widget == nullptr)
+            return "None";
+
+        const auto className = widget->metaObject()->className();
+
+        const auto widgetName = QString("%1:%2")
+                .arg(widget->objectName(), className);
+
+        const auto window = widget->window();
+        if (window && widget != window) {
+            return widgetName
+                + QString(" in %1:%2")
+                    .arg(window->objectName(), window->metaObject()->className());
+        }
+
+        return widgetName;
+    }
+
+    QWidget *keyClicksTarget()
+    {
+        auto popup = QApplication::activePopupWidget();
+        if (popup)
+            return popup;
+
+        auto widget = QApplication::focusWidget();
+        if (widget)
+            return widget;
+
+        auto window = QApplication::activeWindow();
+        if (window)
+            return window->focusWidget();
+
+        auto modal = QApplication::activeModalWidget();
+        if (modal)
+            return modal->focusWidget();
+
+#ifdef Q_OS_MAC
+        return m_wnd->focusWidget();
+#else
+        return nullptr;
+#endif
+    }
+
+
+    MainWindow *m_wnd = nullptr;
+    bool m_succeeded = true;
+    bool m_failed = false;
+};
+#endif // HAS_TESTS
+
 ScriptableProxy::ScriptableProxy(MainWindow *mainWindow, QObject *parent)
     : QObject(parent)
     , m_wnd(mainWindow)
-    , m_tabName()
 {
     qRegisterMetaType< QPointer<QWidget> >("QPointer<QWidget>");
     qRegisterMetaTypeStreamOperators<ClipboardMode>("ClipboardMode");
@@ -644,9 +827,23 @@ ScriptableProxy::ScriptableProxy(MainWindow *mainWindow, QObject *parent)
     qRegisterMetaTypeStreamOperators<Qt::KeyboardModifiers>("Qt::KeyboardModifiers");
 }
 
-QByteArray ScriptableProxy::callFunction(const QByteArray &serializedFunctionCall)
+void ScriptableProxy::callFunction(const QByteArray &serializedFunctionCall)
 {
-    QVector<QVariant> functionCall;
+    auto t = new QTimer(m_wnd);
+    t->setSingleShot(true);
+    QObject::connect( t, &QTimer::timeout, this, [=]() {
+        const auto result = callFunctionHelper(serializedFunctionCall);
+        emit sendMessage(result, CommandFunctionCallReturnValue);
+        t->deleteLater();
+    });
+    t->start(0);
+}
+
+QByteArray ScriptableProxy::callFunctionHelper(const QByteArray &serializedFunctionCall)
+{
+    QVector<QVariant> arguments;
+    QByteArray slotName;
+    int functionCallId;
     {
         QDataStream stream(serializedFunctionCall);
         stream.setVersion(QDataStream::Qt_5_0);
@@ -672,7 +869,21 @@ QByteArray ScriptableProxy::callFunction(const QByteArray &serializedFunctionCal
             return QByteArray();
         }
 
-        stream >> functionCall;
+        stream >> functionCallId;
+        if (stream.status() != QDataStream::Ok) {
+            log("Failed to read scriptable proxy slot call ID", LogError);
+            Q_ASSERT(false);
+            return QByteArray();
+        }
+
+        stream >> slotName;
+        if (stream.status() != QDataStream::Ok) {
+            log("Failed to read scriptable proxy slot call name", LogError);
+            Q_ASSERT(false);
+            return QByteArray();
+        }
+
+        stream >> arguments;
         if (stream.status() != QDataStream::Ok) {
             log("Failed to read scriptable proxy slot call", LogError);
             Q_ASSERT(false);
@@ -680,8 +891,6 @@ QByteArray ScriptableProxy::callFunction(const QByteArray &serializedFunctionCal
         }
     }
 
-    m_tabName = functionCall.value(0).toString();
-    const auto slotName = functionCall.value(1).toByteArray();
     const auto slotIndex = metaObject()->indexOfSlot(slotName);
     if (slotIndex == -1) {
         log("Failed to find scriptable proxy slot: " + slotName, LogError);
@@ -689,75 +898,89 @@ QByteArray ScriptableProxy::callFunction(const QByteArray &serializedFunctionCal
         return QByteArray();
     }
 
-    const int argumentsStartIndex = 2;
-
     const auto metaMethod = metaObject()->method(slotIndex);
     const auto typeId = metaMethod.returnType();
 
     QGenericArgument args[9];
-    for (int i = argumentsStartIndex; i < functionCall.size(); ++i) {
-        auto &value = functionCall[i];
-        const int j = i - argumentsStartIndex;
-        const int argumentTypeId = metaMethod.parameterType(j);
+    for (int i = 0; i < arguments.size(); ++i) {
+        auto &value = arguments[i];
+        const int argumentTypeId = metaMethod.parameterType(i);
         if (argumentTypeId == QMetaType::QVariant) {
-            args[j] = Q_ARG(QVariant, value);
+            args[i] = Q_ARG(QVariant, value);
         } else if ( value.userType() == argumentTypeId ) {
-            args[j] = QGenericArgument( value.typeName(), static_cast<void*>(value.data()) );
+            args[i] = QGenericArgument( value.typeName(), static_cast<void*>(value.data()) );
         } else {
             log( QString("Bad argument type (at index %1) for scriptable proxy slot: %2")
-                 .arg(j)
+                 .arg(i)
                  .arg(metaMethod.methodSignature().constData()), LogError);
             Q_ASSERT(false);
             return QByteArray();
         }
     }
 
+    QVariant returnValue;
+    bool called;
+
     if (typeId == QMetaType::Void) {
-        const bool called = metaMethod.invoke(
+        called = metaMethod.invoke(
                 this, args[0], args[1], args[2], args[3], args[4], args[5], args[6], args[7], args[8]);
+    } else {
+        returnValue = QVariant(typeId, nullptr);
+        const auto genericReturnValue = returnValue.isValid()
+                ? QGenericReturnArgument(returnValue.typeName(), static_cast<void*>(returnValue.data()) )
+                : Q_RETURN_ARG(QVariant, returnValue);
 
-        if (!called) {
-            log( QString("Bad scriptable proxy slot call: %1")
-                 .arg(metaMethod.methodSignature().constData()), LogError);
-            Q_ASSERT(false);
-        }
-
-        return QByteArray();
-    }
-
-    QVariant returnValue(typeId, nullptr);
-    const auto genericReturnValue = returnValue.isValid()
-            ? QGenericReturnArgument(returnValue.typeName(), static_cast<void*>(returnValue.data()) )
-            : Q_RETURN_ARG(QVariant, returnValue);
-
-    const bool called = metaMethod.invoke(
+        called = metaMethod.invoke(
                 this, genericReturnValue,
                 args[0], args[1], args[2], args[3], args[4], args[5], args[6], args[7], args[8]);
+    }
 
     if (!called) {
-        log("Bad scriptable proxy slot call (with return value)", LogError);
+        log( QString("Bad scriptable proxy slot call: %1")
+             .arg(metaMethod.methodSignature().constData()), LogError);
         Q_ASSERT(false);
-        return QByteArray();
     }
 
     QByteArray bytes;
     {
         QDataStream stream(&bytes, QIODevice::WriteOnly);
-        stream << returnValue;
+        stream << functionCallId << returnValue;
     }
 
     return bytes;
 }
 
-void ScriptableProxy::setReturnValue(const QByteArray &returnValue)
+void ScriptableProxy::setFunctionCallReturnValue(const QByteArray &bytes)
 {
-    QDataStream stream(returnValue);
-    stream >> m_returnValue;
+    QDataStream stream(bytes);
+    int functionCallId;
+    QVariant returnValue;
+    stream >> functionCallId >> returnValue;
+    if (stream.status() != QDataStream::Ok) {
+        log("Failed to read scriptable proxy slot call return value", LogError);
+        Q_ASSERT(false);
+        return;
+    }
+    emit functionCallFinished(functionCallId, returnValue);
+}
+
+void ScriptableProxy::setInputDialogResult(const QByteArray &bytes)
+{
+    QDataStream stream(bytes);
+    int dialogId;
+    NamedValueList result;
+    stream >> dialogId >> result;
+    if (stream.status() != QDataStream::Ok) {
+        log("Failed to read input dialog result", LogError);
+        Q_ASSERT(false);
+        return;
+    }
+    emit inputDialogFinished(dialogId, result);
 }
 
 QVariantMap ScriptableProxy::getActionData(int id)
 {
-    INVOKE(getActionData, (id));
+    INVOKE_NO_SNIP(getActionData, (id));
     m_actionData = m_wnd->actionData(id);
     m_actionId = id;
 
@@ -769,7 +992,7 @@ QVariantMap ScriptableProxy::getActionData(int id)
 
 void ScriptableProxy::setActionData(int id, const QVariantMap &data)
 {
-    INVOKE2(setActionData, (id, data));
+    INVOKE_NO_SNIP2(setActionData, (id, data));
     m_wnd->setActionData(id, data);
 }
 
@@ -823,19 +1046,19 @@ bool ScriptableProxy::copyFromCurrentWindow()
 
 bool ScriptableProxy::isMonitoringEnabled()
 {
-    INVOKE(isMonitoringEnabled, ());
+    INVOKE_NO_SNIP(isMonitoringEnabled, ());
     return m_wnd->isMonitoringEnabled();
 }
 
 bool ScriptableProxy::isMainWindowVisible()
 {
-    INVOKE(isMainWindowVisible, ());
+    INVOKE_NO_SNIP(isMainWindowVisible, ());
     return !m_wnd->isMinimized() && m_wnd->isVisible();
 }
 
 bool ScriptableProxy::isMainWindowFocused()
 {
-    INVOKE(isMainWindowFocused, ());
+    INVOKE_NO_SNIP(isMainWindowFocused, ());
     return m_wnd->isActiveWindow();
 }
 
@@ -887,7 +1110,7 @@ QString ScriptableProxy::removeTab(const QString &arg1)
 
 QString ScriptableProxy::tabIcon(const QString &tabName)
 {
-    INVOKE(tabIcon, (tabName));
+    INVOKE_NO_SNIP(tabIcon, (tabName));
     return getIconNameForTabName(tabName);
 }
 
@@ -913,12 +1136,6 @@ bool ScriptableProxy::showBrowserAt(const QString &tabName, QRect rect)
     return showBrowser(tabName);
 }
 
-bool ScriptableProxy::showCurrentBrowser()
-{
-    INVOKE(showCurrentBrowser, ());
-    return showBrowser(m_tabName);
-}
-
 void ScriptableProxy::action(const QVariantMap &arg1, const Command &arg2)
 {
     INVOKE2(action, (arg1, arg2));
@@ -951,10 +1168,10 @@ void ScriptableProxy::showMessage(const QString &title,
     notification->setButtons(buttons);
 }
 
-QVariantMap ScriptableProxy::nextItem(int where)
+QVariantMap ScriptableProxy::nextItem(const QString &tabName, int where)
 {
-    INVOKE(nextItem, (where));
-    ClipboardBrowser *c = fetchBrowser();
+    INVOKE(nextItem, (tabName, where));
+    ClipboardBrowser *c = fetchBrowser(tabName);
     if (!c)
         return QVariantMap();
 
@@ -968,23 +1185,23 @@ QVariantMap ScriptableProxy::nextItem(int where)
     return c->copyIndex(index);
 }
 
-void ScriptableProxy::browserMoveToClipboard(int row)
+void ScriptableProxy::browserMoveToClipboard(const QString &tabName, int row)
 {
-    INVOKE2(browserMoveToClipboard, (row));
-    ClipboardBrowser *c = fetchBrowser();
+    INVOKE2(browserMoveToClipboard, (tabName, row));
+    ClipboardBrowser *c = fetchBrowser(tabName);
     m_wnd->moveToClipboard(c, row);
 }
 
-void ScriptableProxy::browserSetCurrent(int arg1)
+void ScriptableProxy::browserSetCurrent(const QString &tabName, int arg1)
 {
-    INVOKE2(browserSetCurrent, (arg1));
-    BROWSER(setCurrent(arg1));
+    INVOKE2(browserSetCurrent, (tabName, arg1));
+    BROWSER(tabName, setCurrent(arg1));
 }
 
-QString ScriptableProxy::browserRemoveRows(QVector<int> rows)
+QString ScriptableProxy::browserRemoveRows(const QString &tabName, QVector<int> rows)
 {
-    INVOKE(browserRemoveRows, (rows));
-    ClipboardBrowser *c = fetchBrowser();
+    INVOKE(browserRemoveRows, (tabName, rows));
+    ClipboardBrowser *c = fetchBrowser(tabName);
     if (!c)
         return QString("Invalid tab");
 
@@ -1015,16 +1232,16 @@ QString ScriptableProxy::browserRemoveRows(QVector<int> rows)
     return QString();
 }
 
-void ScriptableProxy::browserEditRow(int arg1)
+void ScriptableProxy::browserEditRow(const QString &tabName, int arg1)
 {
-    INVOKE2(browserEditRow, (arg1));
-    BROWSER(editRow(arg1));
+    INVOKE2(browserEditRow, (tabName, arg1));
+    BROWSER(tabName, editRow(arg1));
 }
 
-void ScriptableProxy::browserEditNew(const QString &arg1, bool changeClipboard)
+void ScriptableProxy::browserEditNew(const QString &tabName, const QString &arg1, bool changeClipboard)
 {
-    INVOKE2(browserEditNew, (arg1, changeClipboard));
-    BROWSER(editNew(arg1, changeClipboard));
+    INVOKE2(browserEditNew, (tabName, arg1, changeClipboard));
+    BROWSER(tabName, editNew(arg1, changeClipboard));
 }
 
 QStringList ScriptableProxy::tabs()
@@ -1069,10 +1286,10 @@ bool ScriptableProxy::loadTab(const QString &arg1)
     return m_wnd->loadTab(arg1);
 }
 
-bool ScriptableProxy::saveTab(const QString &arg1)
+bool ScriptableProxy::saveTab(const QString &tabName, const QString &arg1)
 {
-    INVOKE(saveTab, (arg1));
-    ClipboardBrowser *c = fetchBrowser();
+    INVOKE(saveTab, (tabName, arg1));
+    ClipboardBrowser *c = fetchBrowser(tabName);
     if (!c)
         return false;
 
@@ -1136,25 +1353,25 @@ bool ScriptableProxy::hasClipboardFormat(const QString &mime, ClipboardMode mode
     return data && data->hasFormat(mime);
 }
 
-int ScriptableProxy::browserLength()
+int ScriptableProxy::browserLength(const QString &tabName)
 {
-    INVOKE(browserLength, ());
-    ClipboardBrowser *c = fetchBrowser();
+    INVOKE_NO_SNIP(browserLength, (tabName));
+    ClipboardBrowser *c = fetchBrowser(tabName);
     return c ? c->length() : 0;
 }
 
-bool ScriptableProxy::browserOpenEditor(const QByteArray &arg1, bool changeClipboard)
+bool ScriptableProxy::browserOpenEditor(const QString &tabName, const QByteArray &arg1, bool changeClipboard)
 {
-    INVOKE(browserOpenEditor, (arg1, changeClipboard));
-    ClipboardBrowser *c = fetchBrowser();
+    INVOKE(browserOpenEditor, (tabName, arg1, changeClipboard));
+    ClipboardBrowser *c = fetchBrowser(tabName);
     return c && c->openEditor(arg1, changeClipboard);
 }
 
-QString ScriptableProxy::browserInsert(int row, const QVector<QVariantMap> &items)
+QString ScriptableProxy::browserInsert(const QString &tabName, int row, const QVector<QVariantMap> &items)
 {
-    INVOKE(browserInsert, (row, items));
+    INVOKE(browserInsert, (tabName, row, items));
 
-    ClipboardBrowser *c = fetchBrowser();
+    ClipboardBrowser *c = fetchBrowser(tabName);
     if (!c)
         return "Invalid tab";
 
@@ -1169,10 +1386,10 @@ QString ScriptableProxy::browserInsert(int row, const QVector<QVariantMap> &item
     return QString();
 }
 
-bool ScriptableProxy::browserChange(const QVariantMap &data, int row)
+bool ScriptableProxy::browserChange(const QString &tabName, const QVariantMap &data, int row)
 {
-    INVOKE(browserChange, (data, row));
-    ClipboardBrowser *c = fetchBrowser();
+    INVOKE(browserChange, (tabName, data, row));
+    ClipboardBrowser *c = fetchBrowser(tabName);
     if (!c)
         return false;
 
@@ -1188,16 +1405,16 @@ bool ScriptableProxy::browserChange(const QVariantMap &data, int row)
     return c->model()->setData(index, itemData, contentType::data);
 }
 
-QByteArray ScriptableProxy::browserItemData(int arg1, const QString &arg2)
+QByteArray ScriptableProxy::browserItemData(const QString &tabName, int arg1, const QString &arg2)
 {
-    INVOKE(browserItemData, (arg1, arg2));
-    return itemData(arg1, arg2);
+    INVOKE(browserItemData, (tabName, arg1, arg2));
+    return itemData(tabName, arg1, arg2);
 }
 
-QVariantMap ScriptableProxy::browserItemData(int arg1)
+QVariantMap ScriptableProxy::browserItemData(const QString &tabName, int arg1)
 {
-    INVOKE(browserItemData, (arg1));
-    return itemData(arg1);
+    INVOKE(browserItemData, (tabName, arg1));
+    return itemData(tabName, arg1);
 }
 
 void ScriptableProxy::setCurrentTab(const QString &tabName)
@@ -1208,31 +1425,26 @@ void ScriptableProxy::setCurrentTab(const QString &tabName)
         m_wnd->setCurrentTab(c);
 }
 
-void ScriptableProxy::setTab(const QString &tabName)
+QString ScriptableProxy::tab(const QString &tabName)
 {
-    m_tabName = tabName;
-}
-
-QString ScriptableProxy::tab()
-{
-    INVOKE(tab, ());
-    ClipboardBrowser *c = fetchBrowser();
+    INVOKE(tab, (tabName));
+    ClipboardBrowser *c = fetchBrowser(tabName);
     return c ? c->tabName() : QString();
 }
 
 int ScriptableProxy::currentItem()
 {
-    INVOKE(currentItem, ());
+    INVOKE_NO_SNIP(currentItem, ());
 
     const QPersistentModelIndex current =
             m_actionData.value(mimeCurrentItem).value<QPersistentModelIndex>();
     return current.isValid() ? current.row() : -1;
 }
 
-bool ScriptableProxy::selectItems(const QVector<int> &rows)
+bool ScriptableProxy::selectItems(const QString &tabName, const QVector<int> &rows)
 {
-    INVOKE(selectItems, (rows));
-    ClipboardBrowser *c = fetchBrowser();
+    INVOKE(selectItems, (tabName, rows));
+    ClipboardBrowser *c = fetchBrowser(tabName);
     if (!c)
         return false;
 
@@ -1253,7 +1465,7 @@ bool ScriptableProxy::selectItems(const QVector<int> &rows)
 
 QVector<int> ScriptableProxy::selectedItems()
 {
-    INVOKE(selectedItems, ());
+    INVOKE_NO_SNIP(selectedItems, ());
 
     QVector<int> selectedRows;
     const QList<QPersistentModelIndex> selected = selectedIndexes();
@@ -1266,12 +1478,6 @@ QVector<int> ScriptableProxy::selectedItems()
     return selectedRows;
 }
 
-int ScriptableProxy::selectedItemsDataCount()
-{
-    INVOKE(selectedItemsDataCount, ());
-    return selectedIndexes().size();
-}
-
 QVariantMap ScriptableProxy::selectedItemData(int selectedIndex)
 {
     INVOKE(selectedItemData, (selectedIndex));
@@ -1347,17 +1553,23 @@ void ScriptableProxy::setSelectedItemsData(const QVector<QVariantMap> &dataList)
 }
 
 #ifdef HAS_TESTS
-void ScriptableProxy::sendKeys(const QString &keys, int delay)
+void ScriptableProxy::sendKeys(const QString &expectedWidgetName, const QString &keys, int delay)
 {
-    INVOKE2(sendKeys, (keys, delay));
-    m_sentKeyClicks = m_wnd->sendKeyClicks(keys, delay);
+    INVOKE2(sendKeys, (expectedWidgetName, keys, delay));
+    Q_ASSERT( keyClicker()->succeeded() || keyClicker()->failed() );
+    keyClicker()->sendKeyClicks(expectedWidgetName, keys, delay, 10);
 }
 
-bool ScriptableProxy::keysSent()
+bool ScriptableProxy::sendKeysSucceeded()
 {
-    INVOKE(keysSent, ());
-    QCoreApplication::processEvents();
-    return m_wnd->lastReceivedKeyClicks() >= m_sentKeyClicks;
+    INVOKE(sendKeysSucceeded, ());
+    return keyClicker()->succeeded();
+}
+
+bool ScriptableProxy::sendKeysFailed()
+{
+    INVOKE(sendKeysFailed, ());
+    return keyClicker()->failed();
 }
 
 QString ScriptableProxy::testSelected()
@@ -1405,12 +1617,14 @@ QString ScriptableProxy::currentWindowTitle()
     return window ? window->getTitle() : QString();
 }
 
-NamedValueList ScriptableProxy::inputDialog(const NamedValueList &values)
+int ScriptableProxy::inputDialog(const NamedValueList &values)
 {
     INVOKE(inputDialog, (values));
 
-    InputDialog inputDialog(m_wnd);
-    QDialog &dialog = inputDialog.dialog;
+    InputDialog inputDialog;
+    const auto dialogPtr = std::make_shared<QDialog>(m_wnd);
+    inputDialog.dialog = dialogPtr.get();
+    QDialog &dialog = *dialogPtr;
 
     QIcon icon;
     QVBoxLayout layout(&dialog);
@@ -1457,11 +1671,11 @@ NamedValueList ScriptableProxy::inputDialog(const NamedValueList &values)
     if ( !styleSheet.isEmpty() )
         dialog.setStyleSheet(styleSheet);
 
-    QDialogButtonBox buttons(
+    auto buttons = new QDialogButtonBox(
                 QDialogButtonBox::Ok | QDialogButtonBox::Cancel, Qt::Horizontal, &dialog);
-    QObject::connect( &buttons, SIGNAL(accepted()), &dialog, SLOT(accept()) );
-    QObject::connect( &buttons, SIGNAL(rejected()), &dialog, SLOT(reject()) );
-    layout.addWidget(&buttons);
+    QObject::connect( buttons, SIGNAL(accepted()), &dialog, SLOT(accept()) );
+    QObject::connect( buttons, SIGNAL(rejected()), &dialog, SLOT(reject()) );
+    layout.addWidget(buttons);
 
     installShortcutToCloseDialog(&dialog, &dialog, Qt::CTRL | Qt::Key_Enter);
     installShortcutToCloseDialog(&dialog, &dialog, Qt::CTRL | Qt::Key_Return);
@@ -1470,23 +1684,39 @@ NamedValueList ScriptableProxy::inputDialog(const NamedValueList &values)
         icon = appIcon();
     dialog.setWindowIcon(icon);
 
+    int dialogId = ++m_lastInputDialogId;
+    connect(&dialog, &QDialog::finished, this, [=]() {
+        NamedValueList result;
+        result.reserve( widgets.size() );
+
+        if ( dialogPtr->result() ) {
+            for ( auto w : widgets ) {
+                const QString propertyName = w->property(propertyWidgetProperty).toString();
+                const QString name = w->property(propertyWidgetName).toString();
+                const QVariant value = w->property(propertyName.toUtf8().constData());
+                result.append( NamedValue(name, value) );
+            }
+        }
+
+        QByteArray bytes;
+        {
+            QDataStream stream(&bytes, QIODevice::WriteOnly);
+            stream << dialogId << result;
+        }
+
+        emit sendMessage(bytes, CommandInputDialogFinished);
+    });
+
+    // Connecting this directly to QEventLoop::quit() doesn't seem to work always.
+    connect(this, &ScriptableProxy::clientDisconnected, &dialog, &QDialog::reject);
+
     dialog.show();
-    raiseWindow(&dialog);
 
-    if ( !dialog.exec() )
-        return NamedValueList();
+    // Skip raising dialog in tests.
+    if ( !qApp->property("CopyQ_test_id").isValid() )
+        raiseWindow(&dialog);
 
-    NamedValueList result;
-    result.reserve( widgets.size() );
-
-    for ( auto w : widgets ) {
-        const QString propertyName = w->property(propertyWidgetProperty).toString();
-        const QString name = w->property(propertyWidgetName).toString();
-        const QVariant value = w->property(propertyName.toUtf8().constData());
-        result.append( NamedValue(name, value) );
-    }
-
-    return result;
+    return dialogId;
 }
 
 void ScriptableProxy::setUserValue(const QString &key, const QVariant &value)
@@ -1499,7 +1729,7 @@ void ScriptableProxy::setUserValue(const QString &key, const QVariant &value)
 
 void ScriptableProxy::setSelectedItemsData(const QString &mime, const QVariant &value)
 {
-    INVOKE2(setSelectedItemsData, (mime, value));
+    INVOKE_NO_SNIP2(setSelectedItemsData, (mime, value));
     const QList<QPersistentModelIndex> selected = selectedIndexes();
     for (const auto &index : selected) {
         ClipboardBrowser *c = m_wnd->browserForItem(index);
@@ -1520,9 +1750,15 @@ void ScriptableProxy::filter(const QString &text)
     m_wnd->setFilter(text);
 }
 
+QString ScriptableProxy::filter()
+{
+    INVOKE(filter, ());
+    return m_wnd->filter();
+}
+
 QVector<Command> ScriptableProxy::commands()
 {
-    INVOKE(commands, ());
+    INVOKE_NO_SNIP(commands, ());
     return loadAllCommands();
 }
 
@@ -1618,25 +1854,25 @@ Qt::KeyboardModifiers ScriptableProxy::queryKeyboardModifiers()
 
 QString ScriptableProxy::pluginsPath()
 {
-    INVOKE(pluginsPath, ());
+    INVOKE_NO_SNIP(pluginsPath, ());
     return ::pluginsPath();
 }
 
 QString ScriptableProxy::themesPath()
 {
-    INVOKE(themesPath, ());
+    INVOKE_NO_SNIP(themesPath, ());
     return ::themesPath();
 }
 
 QString ScriptableProxy::translationsPath()
 {
-    INVOKE(translationsPath, ());
+    INVOKE_NO_SNIP(translationsPath, ());
     return ::translationsPath();
 }
 
 QString ScriptableProxy::iconColor()
 {
-    INVOKE(iconColor, ());
+    INVOKE_NO_SNIP(iconColor, ());
     const auto color = m_wnd->sessionIconColor();
     return color.isValid() ? color.name() : QString();
 }
@@ -1655,7 +1891,7 @@ bool ScriptableProxy::setIconColor(const QString &colorName)
 
 QString ScriptableProxy::iconTag()
 {
-    INVOKE(iconTag, ());
+    INVOKE_NO_SNIP(iconTag, ());
     return m_wnd->sessionIconTag();
 }
 
@@ -1718,13 +1954,13 @@ void ScriptableProxy::setTitleForData(const QVariantMap &data)
     setTitle(clipboardContent);
 }
 
-void ScriptableProxy::saveData(const QString &tab, const QVariantMap &data)
+void ScriptableProxy::saveData(const QString &tab, const QVariantMap &data, ClipboardMode mode)
 {
-    INVOKE2(saveData, (tab, data));
+    INVOKE2(saveData, (tab, data, mode));
 
     auto c = m_wnd->tab(tab);
     if (c) {
-        c->addUnique(data);
+        c->addUnique(data, mode);
         c->setCurrent(0);
     }
 }
@@ -1790,38 +2026,38 @@ void ScriptableProxy::showDataNotification(const QVariantMap &data)
 
 QStringList ScriptableProxy::menuItemMatchCommands(int actionId)
 {
-    INVOKE(menuItemMatchCommands, (actionId));
+    INVOKE_NO_SNIP(menuItemMatchCommands, (actionId));
     return m_wnd->menuItemMatchCommands(actionId);
 }
 
 bool ScriptableProxy::enableMenuItem(int actionId, int menuItemMatchCommandIndex, bool enabled)
 {
-    INVOKE(enableMenuItem, (actionId, menuItemMatchCommandIndex, enabled));
+    INVOKE_NO_SNIP(enableMenuItem, (actionId, menuItemMatchCommandIndex, enabled));
     return m_wnd->setMenuItemEnabled(actionId, menuItemMatchCommandIndex, enabled);
 }
 
 QVariantMap ScriptableProxy::setDisplayData(int actionId, const QVariantMap &displayData)
 {
-    INVOKE(setDisplayData, (actionId, displayData));
+    INVOKE_NO_SNIP(setDisplayData, (actionId, displayData));
     m_actionData = m_wnd->setDisplayData(actionId, displayData);
     return m_actionData;
 }
 
 QVector<Command> ScriptableProxy::automaticCommands()
 {
-    INVOKE(automaticCommands, ());
+    INVOKE_NO_SNIP(automaticCommands, ());
     return m_wnd->automaticCommands();
 }
 
 QVector<Command> ScriptableProxy::displayCommands()
 {
-    INVOKE(displayCommands, ());
+    INVOKE_NO_SNIP(displayCommands, ());
     return m_wnd->displayCommands();
 }
 
 QVector<Command> ScriptableProxy::scriptCommands()
 {
-    INVOKE(scriptCommands, ());
+    INVOKE_NO_SNIP(scriptCommands, ());
     return m_wnd->scriptCommands();
 }
 
@@ -1848,17 +2084,15 @@ ClipboardBrowser *ScriptableProxy::fetchBrowser(const QString &tabName)
     return tabName.isEmpty() ? m_wnd->browser(0) : m_wnd->tab(tabName);
 }
 
-ClipboardBrowser *ScriptableProxy::fetchBrowser() { return fetchBrowser(m_tabName); }
-
-QVariantMap ScriptableProxy::itemData(int i)
+QVariantMap ScriptableProxy::itemData(const QString &tabName, int i)
 {
-    auto c = fetchBrowser();
+    auto c = fetchBrowser(tabName);
     return c ? c->copyIndex( c->index(i) ) : QVariantMap();
 }
 
-QByteArray ScriptableProxy::itemData(int i, const QString &mime)
+QByteArray ScriptableProxy::itemData(const QString &tabName, int i, const QString &mime)
 {
-    const QVariantMap data = itemData(i);
+    const QVariantMap data = itemData(tabName, i);
     if ( data.isEmpty() )
         return QByteArray();
 
@@ -1892,6 +2126,34 @@ QList<QPersistentModelIndex> ScriptableProxy::selectedIndexes() const
             .value< QList<QPersistentModelIndex> >();
 }
 
+QVariant ScriptableProxy::waitForFunctionCallFinished(int functionCallId)
+{
+    QVariant result;
+
+    QEventLoop loop;
+    connect(this, &ScriptableProxy::functionCallFinished, &loop,
+            [&](int receivedFunctionCallId, const QVariant &returnValue) {
+                if (receivedFunctionCallId != functionCallId)
+                    return;
+                result = returnValue;
+                loop.quit();
+            });
+    connect(this, &ScriptableProxy::clientDisconnected, &loop, &QEventLoop::quit);
+    connect(qApp, &QCoreApplication::aboutToQuit, &loop, &QEventLoop::quit);
+    loop.exec();
+
+    return result;
+}
+
+#ifdef HAS_TESTS
+KeyClicker *ScriptableProxy::keyClicker()
+{
+    if (!m_keyClicker)
+        m_keyClicker = new KeyClicker(m_wnd, this);
+    return m_keyClicker;
+}
+#endif // HAS_TESTS
+
 QString pluginsPath()
 {
     QDir dir;
