@@ -32,10 +32,12 @@
 #include <QVariant>
 #include <QVector>
 
-#include <memory>
+#include <unordered_map>
 
 class ClipboardBrowser;
+class KeyClicker;
 class MainWindow;
+class QEventLoop;
 class QPersistentModelIndex;
 class QPixmap;
 
@@ -78,16 +80,16 @@ class ScriptableProxy : public QObject
     Q_OBJECT
 
 public:
-    /** Create proxy object and move it to same thread as @a mainWindow. */
-    ScriptableProxy(MainWindow* mainWindow, QObject *parent);
+    explicit ScriptableProxy(MainWindow* mainWindow, QObject *parent = nullptr);
 
-    QByteArray callFunction(const QByteArray &serializedFunctionCall);
+    void callFunction(const QByteArray &serializedFunctionCall);
 
     int actionId() const { return m_actionId; }
 
-public slots:
-    void setReturnValue(const QByteArray &returnValue);
+    void setFunctionCallReturnValue(const QByteArray &bytes);
+    void setInputDialogResult(const QByteArray &bytes);
 
+public slots:
     QVariantMap getActionData(int id);
     void setActionData(int id, const QVariantMap &data);
 
@@ -115,8 +117,6 @@ public slots:
     bool showBrowser(const QString &tabName);
     bool showBrowserAt(const QString &tabName, QRect rect);
 
-    bool showCurrentBrowser();
-
     void action(const QVariantMap &arg1, const Command &arg2);
 
     void runInternalAction(const QVariantMap &data, const QString &command);
@@ -128,13 +128,13 @@ public slots:
             const QString &notificationId = QString(),
             const NotificationButtons &buttons = NotificationButtons());
 
-    QVariantMap nextItem(int where);
-    void browserMoveToClipboard(int row);
-    void browserSetCurrent(int arg1);
-    QString browserRemoveRows(QVector<int> rows);
+    QVariantMap nextItem(const QString &tabName, int where);
+    void browserMoveToClipboard(const QString &tabName, int row);
+    void browserSetCurrent(const QString &tabName, int arg1);
+    QString browserRemoveRows(const QString &tabName, QVector<int> rows);
 
-    void browserEditRow(int arg1);
-    void browserEditNew(const QString &arg1, bool changeClipboard);
+    void browserEditRow(const QString &tabName, int arg1);
+    void browserEditNew(const QString &tabName, const QString &arg1, bool changeClipboard);
 
     QStringList tabs();
     bool toggleVisible();
@@ -145,7 +145,7 @@ public slots:
     void openActionDialog(const QVariantMap &arg1);
 
     bool loadTab(const QString &arg1);
-    bool saveTab(const QString &arg1);
+    bool saveTab(const QString &tabName, const QString &arg1);
 
     bool importData(const QString &fileName);
     bool exportData(const QString &fileName);
@@ -156,27 +156,24 @@ public slots:
     QByteArray getClipboardData(const QString &mime, ClipboardMode mode = ClipboardMode::Clipboard);
     bool hasClipboardFormat(const QString &mime, ClipboardMode mode = ClipboardMode::Clipboard);
 
-    int browserLength();
-    bool browserOpenEditor(const QByteArray &arg1, bool changeClipboard);
+    int browserLength(const QString &tabName);
+    bool browserOpenEditor(const QString &tabName, const QByteArray &arg1, bool changeClipboard);
 
-    QString browserInsert(int row, const QVector<QVariantMap> &items);
-    bool browserChange(const QVariantMap &data, int row);
+    QString browserInsert(const QString &tabName, int row, const QVector<QVariantMap> &items);
+    bool browserChange(const QString &tabName, const QVariantMap &data, int row);
 
-    QByteArray browserItemData(int arg1, const QString &arg2);
-    QVariantMap browserItemData(int arg1);
+    QByteArray browserItemData(const QString &tabName, int arg1, const QString &arg2);
+    QVariantMap browserItemData(const QString &tabName, int arg1);
 
     void setCurrentTab(const QString &tabName);
 
-    void setTab(const QString &tabName);
-
-    QString tab();
+    QString tab(const QString &tabName);
 
     int currentItem();
-    bool selectItems(const QVector<int> &rows);
+    bool selectItems(const QString &tabName, const QVector<int> &rows);
 
     QVector<int> selectedItems();
 
-    int selectedItemsDataCount();
     QVariantMap selectedItemData(int selectedIndex);
     bool setSelectedItemData(int selectedIndex, const QVariantMap &data);
 
@@ -184,8 +181,9 @@ public slots:
     void setSelectedItemsData(const QVector<QVariantMap> &dataList);
 
 #ifdef HAS_TESTS
-    void sendKeys(const QString &keys, int delay);
-    bool keysSent();
+    void sendKeys(const QString &expectedWidgetName, const QString &keys, int delay);
+    bool sendKeysSucceeded();
+    bool sendKeysFailed();
     QString testSelected();
 #endif // HAS_TESTS
 
@@ -193,13 +191,14 @@ public slots:
 
     QString currentWindowTitle();
 
-    NamedValueList inputDialog(const NamedValueList &values);
+    int inputDialog(const NamedValueList &values);
 
     void setUserValue(const QString &key, const QVariant &value);
 
     void setSelectedItemsData(const QString &mime, const QVariant &value);
 
     void filter(const QString &text);
+    QString filter();
 
     QVector<Command> commands();
     void setCommands(const QVector<Command> &commands);
@@ -227,7 +226,7 @@ public slots:
     void setClipboardData(const QVariantMap &data);
     void setTitle(const QString &title);
     void setTitleForData(const QVariantMap &data);
-    void saveData(const QString &tab, const QVariantMap &data);
+    void saveData(const QString &tab, const QVariantMap &data, ClipboardMode mode);
     void showDataNotification(const QVariantMap &data);
 
     QStringList menuItemMatchCommands(int actionId);
@@ -242,26 +241,35 @@ public slots:
     bool openUrls(const QStringList &urls);
 
 signals:
-    void sendFunctionCall(const QByteArray &bytes);
+    void functionCallFinished(int functionCallId, const QVariant &returnValue);
+    void inputDialogFinished(int dialogId, const NamedValueList &result);
+    void sendMessage(const QByteArray &message, int messageCode);
+    void clientDisconnected();
 
 private:
     ClipboardBrowser *fetchBrowser(const QString &tabName);
-    ClipboardBrowser *fetchBrowser();
 
-    QVariantMap itemData(int i);
-    QByteArray itemData(int i, const QString &mime);
+    QVariantMap itemData(const QString &tabName, int i);
+    QByteArray itemData(const QString &tabName, int i, const QString &mime);
 
     ClipboardBrowser *currentBrowser() const;
     QList<QPersistentModelIndex> selectedIndexes() const;
 
+    QVariant waitForFunctionCallFinished(int functionId);
+
+    QByteArray callFunctionHelper(const QByteArray &serializedFunctionCall);
+
+#ifdef HAS_TESTS
+    KeyClicker *keyClicker();
+    KeyClicker *m_keyClicker = nullptr;
+#endif // HAS_TESTS
+
     MainWindow* m_wnd;
-    QString m_tabName;
     QVariantMap m_actionData;
     int m_actionId = -1;
 
-    uint m_sentKeyClicks = 0;
-
-    QVariant m_returnValue;
+    int m_lastFunctionCallId = -1;
+    int m_lastInputDialogId = -1;
 };
 
 QString pluginsPath();
