@@ -36,13 +36,13 @@
 #include "gui/tabicons.h"
 #include "platform/platformnativeinterface.h"
 
-#include <QApplication>
 #include <QClipboard>
 #include <QDebug>
 #include <QDir>
 #include <QElapsedTimer>
 #include <QFile>
 #include <QFileInfo>
+#include <QGuiApplication>
 #include <QMap>
 #include <QMimeData>
 #include <QProcess>
@@ -53,18 +53,32 @@
 
 #include <memory>
 
+#define WITH_TIMEOUT "afterMilliseconds(10000, fail); "
+
+// WORKAROUND: Checking clipboard right after closing menu gets stuck on OS X.
+#define ACTIVATE_MENU_ITEM(MENU_ID, WIDGET_ID, CONTENT) \
+    RUN("keys" << MENU_ID << "ENTER", ""); \
+    RUN("keys" << WIDGET_ID, ""); \
+    WAIT_FOR_CLIPBOARD(CONTENT)
+
 namespace {
 
 const auto clipboardTabName = "CLIPBOARD";
 const auto defaultSessionColor = "#ff8800";
 const auto defaultTagColor = "#000000";
 
+const auto clipboardBrowserId = "focus:ClipboardBrowser";
+const auto trayMenuId = "focus:TrayMenu";
+const auto menuId = "focus:Menu";
+const auto editorId = "focus::QTextEdit";
+const auto tabDialogLineEditId = "focus:lineEditTabName";
+
 template <typename Fn1, typename Fn2>
-void runMultiple(Fn1 f1, int intervalMs, Fn2 f2)
+void runMultiple(Fn1 f1, Fn2 f2)
 {
     QTimer timer;
     timer.setSingleShot(true);
-    timer.setInterval(intervalMs);
+    timer.setInterval(0);
     QObject::connect(&timer, &QTimer::timeout, f2);
     timer.start();
     f1();
@@ -80,10 +94,14 @@ bool testStderr(const QByteArray &stderrData, TestInterface::ReadStderrFlag flag
     // Ignore known warnings.
     output.remove("QtWarning: QXcbXSettings::QXcbXSettings(QXcbScreen*) Failed to get selection owner for XSETTINGS_S atom");
     output.remove("QtWarning: QXcbConnection: XCB error:");
+    output.remove("QtWarning: QXcbClipboard: SelectionRequest too old");
     output.remove("QtWarning: libpng warning: iCCP: known incorrect sRGB profile");
     output.remove("QtWarning: QMime::convertToMime: unhandled mimetype: text/plain");
     output.remove("ERROR: QtCritical: QWindowsPipeWriter::write failed. (The pipe is being closed.)");
     output.remove("ERROR: QtCritical: QWindowsPipeWriter: asynchronous write failed. (The pipe has been ended.)");
+#ifdef Q_OS_MAC
+    output.remove("ERROR: Failed to open session mutex: QSystemSemaphore::handle:: ftok failed");
+#endif
 
     if ( output.indexOf(re) != -1 )
         return false;
@@ -97,16 +115,18 @@ bool testStderr(const QByteArray &stderrData, TestInterface::ReadStderrFlag flag
 
 QByteArray getClipboard(const QString &mime = QString("text/plain"))
 {
-    waitFor(waitMsGetClipboard);
-    const QMimeData *data = QApplication::clipboard()->mimeData();
+    const QMimeData *data = clipboardData();
     return (data != nullptr) ? data->data(mime) : QByteArray();
 }
 
 QByteArray waitUntilClipboardSet(const QByteArray &data, const QString &mime = QString("text/plain"))
 {
+    PerformanceTimer perf;
+
     SleepTimer t(waitMsSetClipboard * 5);
     do {
         if (getClipboard(mime) == data) {
+            perf.printPerformance("waitUntilClipboardSet", QStringList() << QString::fromUtf8(data) << mime);
             waitFor(waitMsSetClipboard);
             return data;
         }
@@ -123,30 +143,6 @@ bool waitWhileFileExists(const QFile &file)
     return !file.exists();
 }
 
-bool waitForProcessFinished(QProcess *p)
-{
-    // Process events in case we own clipboard and the new process requests the contents.
-    SleepTimer t(30000);
-    while ( p->state() != QProcess::NotRunning && !p->waitForFinished(50) && t.sleep() ) {}
-    return p->state() == QProcess::NotRunning;
-}
-
-bool closeProcess(QProcess *p)
-{
-    if ( waitForProcessFinished(p) )
-        return true;
-
-    qWarning() << "terminating process";
-    p->terminate();
-
-    if ( !waitForProcessFinished(p) ) {
-        qWarning() << "killing process";
-        terminateProcess(p);
-    }
-
-    return false;
-}
-
 /// Generate unique data.
 QByteArray generateData()
 {
@@ -184,51 +180,34 @@ public:
         if ( isServerRunning() )
             return "Server is already running.";
 
-        if ( isAnyServerRunning() ) {
-            qWarning() << "closing existing test session";
-            run(Args("exit"));
-            waitForAnyServerToQuit();
-        }
-
         m_server.reset(new QProcess);
+        m_serverErrorOutput.clear();
+        QObject::connect(
+            m_server.get(), &QProcess::readyReadStandardError,
+            [this]() {
+                const auto errorOutput = m_server->readAllStandardError().replace('\r', "");
+                m_serverErrorOutput.append(errorOutput);
+            });
 
         if ( !startTestProcess(m_server.get(), QStringList(), QIODevice::ReadOnly) ) {
             return QString("Failed to launch \"%1\": %2")
-                .arg(QApplication::applicationFilePath())
+                .arg(QCoreApplication::applicationFilePath())
                 .arg(m_server->errorString())
                 .toUtf8();
         }
 
-        RETURN_ON_ERROR( waitForServerToStart(), "Failed to start server" );
+        m_server->closeReadChannel(QProcess::StandardOutput);
+
         RETURN_ON_ERROR( readServerErrors(), "Failed to read server errors" );
 
-        return QByteArray();
-    }
-
-    QByteArray waitForServerToStart()
-    {
-        SleepTimer t(15000);
-        do {
-            if ( isServerRunning() )
-                return QByteArray();
-        } while ( t.sleep() );
-
-        return "Unable to start server!" + readServerErrors(ReadAllStderr);
-    }
-
-    QByteArray waitForAnyServerToQuit()
-    {
-        SleepTimer t(8000);
-        do {
-            if ( !isAnyServerRunning() )
-                return QByteArray();
-        } while ( t.sleep() );
-
-        return "Unable to stop server!" + readServerErrors(ReadAllStderr);
+        return waitForServerToStart();
     }
 
     QByteArray stopServer() override
     {
+        if ( !isServerRunning() )
+            return "Server is not running";
+
         QByteArray errors;
         const int exitCode = run(Args("exit"), nullptr, &errors);
         if ( !testStderr(errors) || exitCode != 0 ) {
@@ -236,19 +215,37 @@ public:
                     + printClienAndServerStderr(errors, exitCode);
         }
 
-        if ( m_server != nullptr && !closeProcess(m_server.get()) )
-            return "Failed to close server properly!" + readServerErrors(ReadAllStderr);
+        PerformanceTimer perf;
 
-        return waitForAnyServerToQuit();
+        // Process events in case we own clipboard and the new process requests the contents.
+        SleepTimer t(30000);
+        while ( m_server->state() != QProcess::NotRunning && !m_server->waitForFinished(50) && t.sleep() ) {}
+
+        perf.printPerformance("stopServer");
+
+        if ( m_server->state() != QProcess::NotRunning ) {
+            qWarning() << "terminating server process";
+            m_server->terminate();
+
+            if ( !m_server->waitForFinished() ) {
+                qWarning() << "killing server process";
+                terminateProcess(m_server.get());
+            }
+
+            return "Failed to close server properly!" + readServerErrors(ReadAllStderr);
+        }
+
+        return readServerErrors();
     }
 
     bool isServerRunning() override
     {
-        return m_server != nullptr && m_server->state() == QProcess::Running && isAnyServerRunning();
+        return m_server != nullptr && m_server->state() == QProcess::Running;
     }
 
     int run(const QStringList &arguments, QByteArray *stdoutData = nullptr,
-            QByteArray *stderrData = nullptr, const QByteArray &in = QByteArray()) override
+            QByteArray *stderrData = nullptr, const QByteArray &in = QByteArray(),
+            const QStringList &environment = QStringList()) override
     {
         if (stdoutData != nullptr)
             stdoutData->clear();
@@ -257,7 +254,7 @@ public:
             stderrData->clear();
 
         QProcess p;
-        if (!startTestProcess(&p, arguments))
+        if (!startTestProcess(&p, arguments, QIODevice::ReadWrite, environment))
             return -1;
 
         if ( p.write(in) != in.size() )
@@ -271,6 +268,8 @@ public:
         if (stderrData == nullptr)
             p.closeReadChannel(QProcess::StandardError);
 
+        PerformanceTimer perf;
+
         SleepTimer t(waitClientRun);
         while ( p.state() == QProcess::Running ) {
             if ( stdoutData != nullptr ) {
@@ -284,7 +283,7 @@ public:
             }
 
             if ( !t.sleep() ) {
-                qWarning() << "client process timed out";
+                qWarning() << "Client process timed out" << arguments;
                 return -1;
             }
         }
@@ -299,6 +298,8 @@ public:
             stdoutData->replace('\r', "");
         }
 
+        perf.printPerformance("run", arguments);
+
         return p.exitCode();
     }
 
@@ -374,43 +375,29 @@ public:
 
     QByteArray setClipboard(const QByteArray &bytes, const QString &mime) override
     {
+        if ( getClipboard(mime) == bytes )
+            return QByteArray();
+
         waitFor(waitMsSetClipboard);
 
         auto mimeData = new QMimeData();
         mimeData->setData(mime, bytes);
-        QApplication::clipboard()->setMimeData(mimeData);
+        QGuiApplication::clipboard()->setMimeData(mimeData);
+
+        waitFor(waitMsSetClipboard);
 
         waitUntilClipboardSet(bytes, mime);
         RETURN_ON_ERROR( testClipboard(bytes, mime), "Failed to set clipboard" );
 
-        return "";
+        return QByteArray();
     }
 
     QByteArray readServerErrors(ReadStderrFlag flag = ReadErrors) override
     {
         if (m_server) {
-            QByteArray output = m_server->readAllStandardError();
-
-            // Flush server output.
-            if (m_server->state() != QProcess::NotRunning) {
-                const QByteArray data = "flush ID: " + generateData();
-                if ( run(Args("serverLog") << data) == 0 ) {
-                    SleepTimer t(5000);
-                    bool isFlushed = false;
-                    do {
-                        output.append(m_server->readAllStandardError());
-                        isFlushed = output.contains(data);
-                    } while (!isFlushed && t.sleep());
-
-                    if (!isFlushed)
-                        qWarning() << "failed to flush server output";
-                }
-            }
-
-            output.replace('\r', "");
-
-            if ( flag == ReadAllStderr || !testStderr(output, flag) )
-              return decorateOutput("Server STDERR", output);
+            QCoreApplication::processEvents();
+            if ( flag == ReadAllStderr || !testStderr(m_serverErrorOutput, flag) )
+              return decorateOutput("Server STDERR", m_serverErrorOutput);
         }
 
         return QByteArray();
@@ -430,13 +417,6 @@ public:
         return "";
     }
 
-    QByteArray show() override
-    {
-        const QByteArray out = runClient(Args("show"), "");
-        waitFor(waitMsShow);
-        return out;
-    }
-
     QByteArray cleanupTestCase() override
     {
         return cleanup();
@@ -454,11 +434,21 @@ public:
         if ( isServerRunning() )
             RETURN_ON_ERROR( stopServer(), "Failed to stop server" );
 
-        // Remove all configuration.
+        // Remove all configuration files and tab data.
         const auto settingsPath = settingsDirectoryPath();
+        Q_ASSERT( !settingsPath.isEmpty() );
         QDir settingsDir(settingsPath);
-        if ( settingsDir.exists() && !settingsDir.removeRecursively() )
-            return "Failed to remove settings directory " + settingsPath.toUtf8();
+        const QStringList settingsFileFilters("copyq.test*");
+        // Omit using dangerous QDir::removeRecursively().
+        for ( const auto &fileName : settingsDir.entryList(settingsFileFilters) ) {
+            const auto path = settingsDir.absoluteFilePath(fileName);
+            QFile settingsFile(path);
+            if ( !settingsFile.remove() ) {
+                return QString("Failed to remove settings file \"%1\": %2")
+                    .arg(settingsPath, settingsFile.errorString())
+                    .toUtf8();
+            }
+        }
 
         // Update settings for tests.
         {
@@ -497,7 +487,9 @@ public:
 
         // Always show main window first so that the results are consistent with desktop environments
         // where user cannot hide main window (tiling window managers without tray).
-        return show();
+        RETURN_ON_ERROR( runClient(Args("show"), ""), "Failed to show main window" );
+
+        return QByteArray();
     }
 
     QByteArray cleanup() override
@@ -528,16 +520,25 @@ private:
         QCOMPARE( tabs(), QStringList() );
     }
 
-    bool isAnyServerRunning()
-    {
-        return run(Args("eval")) == 0;
-    }
-
     bool startTestProcess(QProcess *p, const QStringList &arguments,
-                          QIODevice::OpenMode mode = QIODevice::ReadWrite)
+                          QIODevice::OpenMode mode = QIODevice::ReadWrite,
+                          const QStringList &environment = QStringList())
     {
-        p->setProcessEnvironment(m_env);
-        p->start( QApplication::applicationFilePath(), arguments, mode );
+        if ( environment.isEmpty() ) {
+            p->setProcessEnvironment(m_env);
+        } else {
+            auto env = m_env;
+            for (const QString &nameValue : environment) {
+                const auto i = nameValue.indexOf(QLatin1Char('='));
+                Q_ASSERT(i != -1);
+                const auto name = nameValue.left(i);
+                const auto value = nameValue.mid(i+1);
+                env.insert(name, value);
+            }
+            p->setProcessEnvironment(env);
+        }
+
+        p->start( QCoreApplication::applicationFilePath(), arguments, mode );
         return p->waitForStarted(10000);
     }
 
@@ -554,7 +555,19 @@ private:
         return "";
     }
 
+    QByteArray waitForServerToStart()
+    {
+        SleepTimer t(15000);
+        do {
+            if ( run(Args() << "") == 0 )
+                return QByteArray();
+        } while ( t.sleep() );
+
+        return "Unable to start server!" + readServerErrors(ReadAllStderr);
+    }
+
     std::unique_ptr<QProcess> m_server;
+    QByteArray m_serverErrorOutput;
     QProcessEnvironment m_env;
     QString m_testId;
     QVariantMap m_settings;
@@ -640,7 +653,7 @@ void Tests::commandVersion()
     // Version contains application name and version.
     QVERIFY( version.contains(QRegExp("\\bCopyQ\\b.*" + QRegExp::escape(COPYQ_VERSION))) );
     // Version contains Qt version.
-    QVERIFY( version.contains(QRegExp("\\bQt\\s+\\d")) );
+    QVERIFY( version.contains(QRegExp("\\bQt:\\s+\\d")) );
 }
 
 void Tests::badCommand()
@@ -741,11 +754,9 @@ void Tests::commandToggle()
 {
     RUN("visible", "true\n");
     RUN("toggle", "false\n");
-    waitFor(waitMsShow);
     WAIT_ON_OUTPUT("visible", "false\n");
 
     RUN("toggle", "true\n");
-    waitFor(waitMsShow);
     WAIT_ON_OUTPUT("visible", "true\n");
 }
 
@@ -753,7 +764,6 @@ void Tests::commandHide()
 {
     RUN("visible", "true\n");
     RUN("hide", "");
-    waitFor(waitMsShow);
     WAIT_ON_OUTPUT("visible", "false\n");
 }
 
@@ -761,11 +771,9 @@ void Tests::commandShow()
 {
     RUN("visible", "true\n");
     RUN("hide", "");
-    waitFor(waitMsShow);
     WAIT_ON_OUTPUT("visible", "false\n");
 
     RUN("show", "");
-    waitFor(waitMsShow);
     WAIT_ON_OUTPUT("visible", "true\n");
 }
 
@@ -773,11 +781,9 @@ void Tests::commandShowAt()
 {
     RUN("visible", "true\n");
     RUN("hide", "");
-    waitFor(waitMsShow);
     WAIT_ON_OUTPUT("visible", "false\n");
 
     RUN("showAt", "");
-    waitFor(waitMsShow);
     WAIT_ON_OUTPUT("visible", "true\n");
 }
 
@@ -904,41 +910,49 @@ void Tests::commandToggleConfig()
 
 void Tests::commandDialog()
 {
+    RUN(Args() << "keys" << clipboardBrowserId, "");
     runMultiple(
-        [&]() { RUN("dialog" << "text", "TEST\n"); },
-        waitMsShow,
-        [&]() { RUN(Args() << "keys" << ":TEST" << "ENTER", ""); }
+        [&]() { RUN(WITH_TIMEOUT "dialog('text')", "TEST\n"); },
+        [&]() { RUN(Args() << "keys" << "focus::QLineEdit in :QDialog" << ":TEST" << "ENTER", ""); }
     );
 
+    RUN(Args() << "keys" << clipboardBrowserId, "");
     runMultiple(
-        [&]() { RUN("eval" << "dialog('text') === undefined", "true\n"); },
-        waitMsShow,
-        [&]() { RUN(Args() << "keys" << "ESCAPE", ""); }
+        [&]() { RUN(WITH_TIMEOUT "dialog('text') === undefined", "true\n"); },
+        [&]() { RUN(Args() << "keys" << "focus::QLineEdit in :QDialog" << "ESCAPE", ""); }
     );
 
+    RUN(Args() << "keys" << clipboardBrowserId, "");
     runMultiple(
-        [&]() { RUN("eval" << "dialog('.defaultChoice', 2, 'list', [1, 2, 3])", "2\n"); },
-        waitMsShow,
-        [&]() { RUN(Args() << "keys" << "ENTER", ""); }
+        [&]() { RUN(WITH_TIMEOUT "dialog('.defaultChoice', 2, 'list', [1, 2, 3])", "2\n"); },
+        [&]() { RUN(Args() << "keys" << "focus::QComboBox in :QDialog" << "ENTER", ""); }
     );
 
+    RUN(Args() << "keys" << clipboardBrowserId, "");
     runMultiple(
-        [&]() { RUN("eval" << "dialog('.defaultChoice', '', 'list', [1, 2, 3])", "\n"); },
-        waitMsShow,
-        [&]() { RUN(Args() << "keys" << "ENTER", ""); }
+        [&]() { RUN(WITH_TIMEOUT "dialog('.defaultChoice', '', 'list', [1, 2, 3])", "\n"); },
+        [&]() { RUN(Args() << "keys" << "focus::QComboBox in :QDialog" << "ENTER", ""); }
     );
 
+    RUN(Args() << "keys" << clipboardBrowserId, "");
     runMultiple(
-        [&]() { RUN("eval" << "dialog('list', [0, 1, 2])", "0\n"); },
-        waitMsShow,
-        [&]() { RUN(Args() << "keys" << "ENTER", ""); }
+        [&]() { RUN(WITH_TIMEOUT "dialog('list', [0, 1, 2])", "0\n"); },
+        [&]() { RUN(Args() << "keys" << "focus::QComboBox in :QDialog" << "ENTER", ""); }
     );
 
+    // Can't focus configuration checkboxes on OS X
+#ifndef Q_OS_MAC
+    RUN(Args() << "keys" << clipboardBrowserId, "");
     runMultiple(
-        [&]() { RUN("eval" << "dialog('boolean', true) === true", "true\n"); },
-        waitMsShow,
-        [&]() { RUN(Args() << "keys" << "ENTER", ""); }
+        [&]() { RUN(WITH_TIMEOUT "dialog('boolean', true) === true", "true\n"); },
+        [&]() { RUN(Args() << "keys" << "focus::QCheckBox in :QDialog" << "ENTER", ""); }
     );
+#endif
+}
+
+void Tests::commandDialogCloseOnDisconnect()
+{
+    RUN("afterMilliseconds(0, abort); dialog()", "");
 }
 
 void Tests::commandsPackUnpack()
@@ -1364,6 +1378,20 @@ void Tests::commandAfterMilliseconds()
     RUN(script + "sleep(200)", "TEST");
 }
 
+void Tests::commandAsync()
+{
+    RUN("afterMilliseconds(0, function() { print(currentItem()); abort(); }); dialog()", "-1");
+}
+
+void Tests::commandFilter()
+{
+    RUN("filter", "\n");
+    RUN("filter" << "test", "");
+    RUN("filter", "test\n");
+    RUN("filter" << "", "");
+    RUN("filter", "\n");
+}
+
 void Tests::classFile()
 {
     RUN("var f = new File('/copyq_missing_file'); f.exists()", "false\n");
@@ -1542,11 +1570,11 @@ void Tests::keysAndFocusing()
 #endif
 
     RUN("disable", "");
-    RUN("keys" << "CTRL+T", "");
+    RUN("keys" << clipboardBrowserId << "CTRL+T", "");
 
     WAIT_ON_OUTPUT("currentWindowTitle", "CopyQ New Tab\n");
 
-    RUN("keys" << "ESC", "");
+    RUN("keys" << tabDialogLineEditId << "ESC", "");
     WAIT_ON_OUTPUT("currentWindowTitle", "CopyQ\n");
     RUN("enable", "");
 }
@@ -1641,7 +1669,9 @@ void Tests::copyItems()
 void Tests::createTabDialog()
 {
     const auto tab1 = testTab(1);
-    RUN("keys" << "CTRL+T" << ":" + tab1 << "ENTER", "");
+    RUN("keys"
+        << clipboardBrowserId << "CTRL+T"
+        << tabDialogLineEditId << ":" + tab1 << "ENTER", "");
     RUN("testSelected", tab1 + "\n");
 }
 
@@ -1651,11 +1681,14 @@ void Tests::editItems()
 
     RUN("add" << "Line 4" << "Line 1", "");
 
-    RUN("keys" << "F2" << "END" << "ENTER" << ":Line 2" << "F2", "");
+    RUN("keys"
+        << clipboardBrowserId << "F2"
+        << editorId << "END" << "ENTER" << ":Line 2" << "F2", "");
     RUN("read" << "0", "Line 1\nLine 2");
 
-    RUN("keys" << "DOWN", "");
-    RUN("keys" << "F2" << "HOME" << ":Line 3" << "ENTER" << "F2", "");
+    RUN("keys"
+        << clipboardBrowserId << "DOWN" << "F2"
+        << editorId << "HOME" << ":Line 3" << "ENTER" << "F2", "");
     RUN("read" << "1", "Line 3\nLine 4");
     RUN("read" << "0", "Line 1\nLine 2");
 }
@@ -1986,7 +2019,7 @@ void Tests::externalEditor()
                     "add(arguments[1]); while(length()) sleep(100);"
                 )"
                 "--")
-            .arg(QApplication::applicationFilePath())
+            .arg(QCoreApplication::applicationFilePath())
             .arg(editorTab)
             + " %1";
     RUN("config" << "editor" << cmd, cmd + "\n");
@@ -2136,18 +2169,16 @@ void Tests::pasteFromMainWindow()
     RUN("add" << "TEST", "");
     RUN("hide", "");
     runMultiple(
-        [&]() { RUN("dialog" << "text", "TEST\n"); },
-        waitMsShow,
+        [&]() { RUN(WITH_TIMEOUT "dialog('text')", "TEST\n"); },
         [&]() {
+            RUN("keys" << "focus::QLineEdit in :QDialog", "");
             RUN("show", "");
-            waitFor(waitMsShow);
-            RUN("keys" << "ENTER", "");
-            waitFor(waitMsShow);
+            RUN("keys" << clipboardBrowserId << "ENTER", "");
 
             WAIT_FOR_CLIPBOARD("TEST");
             waitFor(waitMsPasteClipboard);
 
-            RUN("keys" << "ENTER", "");
+            RUN("keys" << "focus::QLineEdit in :QDialog" << "ENTER", "");
         }
     );
 }
@@ -2155,11 +2186,9 @@ void Tests::pasteFromMainWindow()
 void Tests::tray()
 {
     RUN("add" << "A", "");
-    waitFor(waitTrayUpdate);
+    RUN("keys" << clipboardBrowserId, "");
     RUN("menu", "");
-    waitFor(waitMsShow);
-    RUN("keys" << "ENTER", "");
-    WAIT_FOR_CLIPBOARD("A");
+    ACTIVATE_MENU_ITEM(trayMenuId, clipboardBrowserId, "A");
 }
 
 void Tests::menu()
@@ -2168,29 +2197,22 @@ void Tests::menu()
 
     RUN("tab" << tab << "add" << "D" << "C" << "B" << "A", "");
 
-    waitFor(waitTrayUpdate);
     RUN("menu" << tab, "");
-    waitFor(waitMsShow);
-    RUN("keys" << "ENTER", "");
-    WAIT_FOR_CLIPBOARD("A");
+    ACTIVATE_MENU_ITEM(menuId, clipboardBrowserId, "A");
 
     // Show menu with 2 items from the tab and select last one.
-    waitFor(waitTrayUpdate);
     RUN("menu" << tab << "2", "");
-    waitFor(waitMsShow);
-    RUN("keys" << "END" << "ENTER", "");
-    WAIT_FOR_CLIPBOARD("B");
+    RUN("keys" << menuId << "END", "");
+    ACTIVATE_MENU_ITEM(menuId, clipboardBrowserId, "B");
 }
 
 void Tests::traySearch()
 {
     RUN("add" << "C" << "B" << "A", "");
 
-    waitFor(waitTrayUpdate);
     RUN("menu", "");
-    waitFor(waitMsShow);
-    RUN("keys" << "B" << "ENTER", "");
-    WAIT_FOR_CLIPBOARD("B");
+    RUN("keys" << trayMenuId << "B", "");
+    ACTIVATE_MENU_ITEM(trayMenuId, clipboardBrowserId, "B");
 }
 
 void Tests::trayPaste()
@@ -2199,30 +2221,28 @@ void Tests::trayPaste()
 
     const auto tab1 = testTab(1);
     RUN("setCurrentTab" << tab1, "");
-    RUN("keys" << "CTRL+N" << ":NEW ", "");
+    RUN("keys"
+        << clipboardBrowserId << "CTRL+N"
+        << editorId << ":NEW ", "");
 
     RUN("add" << "TEST", "");
-    waitFor(waitTrayUpdate);
     RUN("menu", "");
-    waitFor(waitMsShow);
-    RUN("keys" << "ENTER", "");
-    WAIT_FOR_CLIPBOARD("TEST");
+    ACTIVATE_MENU_ITEM(trayMenuId, editorId, "TEST");
     waitFor(waitMsPasteClipboard);
 
-    RUN("keys" << "F2", "");
+    RUN("keys" << editorId << "F2", "");
     RUN("tab" << tab1 << "read" << "0", "NEW TEST");
 
-    RUN("keys" << "CTRL+N" << ":NEW ", "");
+    RUN("keys"
+        << clipboardBrowserId << "CTRL+N"
+        << editorId << ":NEW ", "");
 
     RUN("config" << "tray_item_paste" << "false", "false\n");
     RUN("add" << "TEST2", "");
-    waitFor(waitTrayUpdate);
     RUN("menu", "");
-    waitFor(waitMsShow);
-    RUN("keys" << "ENTER", "");
-    WAIT_FOR_CLIPBOARD("TEST2");
+    ACTIVATE_MENU_ITEM(trayMenuId, editorId, "TEST2");
 
-    RUN("keys" << "F2", "");
+    RUN("keys" << editorId << "F2", "");
     RUN("tab" << tab1 << "read" << "0", "NEW ");
 }
 
@@ -2238,19 +2258,13 @@ void Tests::configTrayTab()
 
     RUN("config" << "tray_tab" << tab1, tab1 + "\n");
 
-    waitFor(waitTrayUpdate);
     RUN("menu", "");
-    waitFor(waitMsShow);
-    RUN("keys" << "ENTER", "");
-    WAIT_FOR_CLIPBOARD("A");
+    ACTIVATE_MENU_ITEM(trayMenuId, clipboardBrowserId, "A");
 
     RUN("config" << "tray_tab" << tab2, tab2 + "\n");
 
-    waitFor(waitTrayUpdate);
     RUN("menu", "");
-    waitFor(waitMsShow);
-    RUN("keys" << "ENTER", "");
-    WAIT_FOR_CLIPBOARD("B");
+    ACTIVATE_MENU_ITEM(trayMenuId, clipboardBrowserId, "B");
 }
 
 void Tests::configMove()
@@ -2263,18 +2277,16 @@ void Tests::configMove()
 
     RUN("config" << "move" << "true", "true\n");
 
-    waitFor(waitTrayUpdate);
     RUN("menu", "");
-    waitFor(waitMsShow);
-    RUN("keys" << "DOWN" << "ENTER", "");
+    RUN("keys" << trayMenuId << "DOWN", "");
+    ACTIVATE_MENU_ITEM(trayMenuId, clipboardBrowserId, "B");
     RUN("read" << "0" << "1", "B\nA");
 
     RUN("config" << "move" << "false", "false\n");
 
-    waitFor(waitTrayUpdate);
     RUN("menu", "");
-    waitFor(waitMsShow);
-    RUN("keys" << "DOWN" << "ENTER", "");
+    RUN("keys" << trayMenuId << "DOWN", "");
+    ACTIVATE_MENU_ITEM(trayMenuId, clipboardBrowserId, "A");
     RUN("read" << "0" << "1", "B\nA");
 }
 
@@ -2289,18 +2301,12 @@ void Tests::configTrayTabIsCurrent()
     RUN("config" << "tray_tab_is_current" << "true", "true\n");
 
     RUN("setCurrentTab" << tab1, "");
-    waitFor(waitTrayUpdate);
     RUN("menu", "");
-    waitFor(waitMsShow);
-    RUN("keys" << "ENTER", "");
-    WAIT_FOR_CLIPBOARD("A");
+    ACTIVATE_MENU_ITEM(trayMenuId, clipboardBrowserId, "A");
 
     RUN("setCurrentTab" << tab2, "");
-    waitFor(waitTrayUpdate);
     RUN("menu", "");
-    waitFor(waitMsShow);
-    RUN("keys" << "ENTER", "");
-    WAIT_FOR_CLIPBOARD("B");
+    ACTIVATE_MENU_ITEM(trayMenuId, clipboardBrowserId, "B");
 }
 
 void Tests::configAutostart()
@@ -2316,6 +2322,20 @@ void Tests::configAutostart()
     RUN("config" << "autostart", "false\n");
 }
 
+void Tests::configPathEnvVariable()
+{
+    const auto path = QDir::home().absoluteFilePath("copyq-settings");
+    const auto environment = QStringList("COPYQ_SETTINGS_PATH=" + path);
+
+    QByteArray out;
+    QByteArray err;
+    run(Args() << "info" << "config", &out, &err, QByteArray(), environment);
+    QVERIFY2( testStderr(err), err );
+
+    const auto expectedOut = path.toUtf8();
+    QCOMPARE( out.left(expectedOut.size()), expectedOut );
+}
+
 void Tests::shortcutCommand()
 {
     RUN("setCommands([{name: 'test', inMenu: true, shortcuts: ['Ctrl+F1'], cmd: 'copyq add OK'}])", "");
@@ -2361,7 +2381,10 @@ void Tests::shortcutCommandMatchInput()
 
 void Tests::shortcutCommandMatchCmd()
 {
-    // Activate only one of the two actions depending on exit code of command which matches input MIME format.
+    const auto tab = testTab(1);
+    const Args args = Args("tab") << tab;
+
+    // Activate only one of the three actions depending on exit code of command which matches input MIME format.
     const auto script = R"(
         function cmd(name) {
           var format = 'application/x-copyq-' + name
@@ -2370,7 +2393,7 @@ void Tests::shortcutCommandMatchCmd()
             inMenu: true,
             shortcuts: ['Ctrl+F1'],
             matchCmd: 'copyq: str(data("' + format + '")) || fail()',
-            cmd: 'copyq add ' + name
+            cmd: 'copyq tab )" + tab + R"( add ' + name
           }
         }
         setCommands([ cmd('test1'), cmd('test2') ])
@@ -2378,14 +2401,10 @@ void Tests::shortcutCommandMatchCmd()
     RUN(script, "");
 
     RUN("write" << "application/x-copyq-test1" << "1", "");
-    waitFor(500);
-    WAIT_ON_OUTPUT("keys('Ctrl+F1'); read(0)", "test1");
-    RUN("tab" << QString(clipboardTabName) << "size", "2\n");
+    WAIT_ON_OUTPUT(args << "keys('Ctrl+F1'); read(0)", "test1");
 
-    RUN("write" << "application/x-copyq-test2" << "1", "");
-    waitFor(500);
-    WAIT_ON_OUTPUT("keys('Ctrl+F1'); read(0)", "test2");
-    RUN("tab" << QString(clipboardTabName) << "size", "4\n");
+    RUN("write" << "application/x-copyq-test2" << "2", "");
+    WAIT_ON_OUTPUT(args << "keys('Ctrl+F1'); read(0)", "test2");
 }
 
 void Tests::shortcutCommandSelectedItemData()
@@ -2406,7 +2425,7 @@ void Tests::shortcutCommandSelectedItemData()
     RUN("add" << "C" << "B" << "A", "");
     RUN("selectItems" << "1" << "2", "true\n");
     RUN("keys" << "CTRL+F1", "");
-    RUN("tab" << tab1 << "read" << "0", "C");
+    WAIT_ON_OUTPUT("tab" << tab1 << "read" << "0", "C");
 }
 
 void Tests::shortcutCommandSetSelectedItemData()
@@ -2424,7 +2443,7 @@ void Tests::shortcutCommandSetSelectedItemData()
     RUN("add" << "C" << "B" << "A", "");
     RUN("selectItems" << "1" << "2", "true\n");
     RUN("keys" << "CTRL+F1", "");
-    RUN("read" << "2", "X");
+    WAIT_ON_OUTPUT("read" << "2", "X");
     RUN("read" << "DATA" << "2", "TEST");
 }
 
@@ -2447,7 +2466,7 @@ void Tests::shortcutCommandSelectedItemsData()
     RUN("add" << "C" << "B" << "A", "");
     RUN("selectItems" << "1" << "2", "true\n");
     RUN("keys" << "CTRL+F1", "");
-    RUN("tab" << tab1 << "read" << "0", "B,C,");
+    WAIT_ON_OUTPUT("tab" << tab1 << "read" << "0", "B,C,");
 }
 
 void Tests::shortcutCommandSetSelectedItemsData()
@@ -2465,9 +2484,7 @@ void Tests::shortcutCommandSetSelectedItemsData()
     RUN("add" << "C" << "B" << "A", "");
     RUN("selectItems" << "1" << "2", "true\n");
     RUN("keys" << "CTRL+F1", "");
-    RUN("read" << "0", "A");
-    RUN("read" << "1", "X");
-    RUN("read" << "2", "Y");
+    WAIT_ON_OUTPUT("read" << "0" << "1" << "2", "A\nX\nY");
 }
 
 void Tests::shortcutCommandSelectedAndCurrent()
@@ -2488,7 +2505,7 @@ void Tests::shortcutCommandSelectedAndCurrent()
 
     RUN("tab" << tab1 << "setCurrentTab" << tab1 << "selectItems" << "1" << "2", "true\n");
     RUN("keys" << "CTRL+F1", "");
-    RUN("tab" << tab1 << "read(0)", "1,2|2|" + tab1);
+    WAIT_ON_OUTPUT("tab" << tab1 << "read(0)", "1,2|2|" + tab1.toUtf8());
 }
 
 void Tests::automaticCommandIgnore()
@@ -2712,9 +2729,11 @@ void Tests::setTabName()
     RUN(script, "1,0");
 }
 
-int Tests::run(const QStringList &arguments, QByteArray *stdoutData, QByteArray *stderrData, const QByteArray &in)
+int Tests::run(
+        const QStringList &arguments, QByteArray *stdoutData, QByteArray *stderrData, const QByteArray &in,
+        const QStringList &environment)
 {
-    return m_test->run(arguments, stdoutData, stderrData, in);
+    return m_test->run(arguments, stdoutData, stderrData, in, environment);
 }
 
 bool Tests::hasTab(const QString &tabName)
@@ -2744,7 +2763,8 @@ int runTests(int argc, char *argv[])
         }
     }
 
-    QApplication app(argc, argv);
+    QGuiApplication app(argc, argv);
+    Q_UNUSED(app);
 
     const QString session = "copyq.test";
     QCoreApplication::setOrganizationName(session);
