@@ -77,6 +77,7 @@ private slots:
     void commandToggleConfig();
 
     void commandDialog();
+    void commandDialogCloseOnDisconnect();
 
     void commandsPackUnpack();
     void commandsBase64();
@@ -124,6 +125,10 @@ private slots:
 
     void commandAfterMilliseconds();
 
+    void commandAsync();
+
+    void commandFilter();
+
     void classFile();
     void classDir();
     void classTemporaryFile();
@@ -185,6 +190,8 @@ private slots:
 
     void configAutostart();
 
+    void configPathEnvVariable();
+
     void shortcutCommand();
     void shortcutCommandOverrideEnter();
     void shortcutCommandMatchInput();
@@ -218,7 +225,8 @@ private slots:
 private:
     void clearServerErrors();
     int run(const QStringList &arguments, QByteArray *stdoutData = nullptr,
-            QByteArray *stderrData = nullptr, const QByteArray &in = QByteArray());
+            QByteArray *stderrData = nullptr, const QByteArray &in = QByteArray(),
+            const QStringList &environment = QStringList());
     bool hasTab(const QString &tabName);
 
     TestInterfacePtr m_test;
